@@ -7,6 +7,8 @@
 /* Date:   01. Jun. 95                                                    */
 /**************************************************************************/
 
+#include "table.hpp"
+
 namespace netgen
 {
 
@@ -149,7 +151,14 @@ public:
     int pos = Position (bnr, ahash);
     return cont.Get (bnr, pos);
   }
-  
+
+  T & Get (const INDEX_2 & ahash)
+  {
+    int bnr = HashValue (ahash);
+    int pos = Position (bnr, ahash);
+    return cont.Get (bnr, pos);
+  }
+
   ///
   bool Used (const INDEX_2 & ahash) const
   {
@@ -214,9 +223,14 @@ public:
     int BagNr() const { return bagnr; }
     int Pos() const { return pos; }
 
-    void operator++ (int)
+    Iterator operator++ (int)
+    {
+      Iterator it(ht, bagnr, pos);
+      ++(*this);
+      return it;
+    }
+    Iterator& operator++()
     {
-      // cout << "begin Operator ++: bagnr = " << bagnr << " -  pos = " << pos << endl;
       pos++;
       while (bagnr < ht.GetNBags() && 
 	     pos == ht.GetBagSize(bagnr+1))
@@ -224,7 +238,12 @@ public:
 	  pos = 0;
 	  bagnr++;
 	}
-      // cout << "end Operator ++: bagnr = " << bagnr << " - pos = " << pos << endl;
+      return *this;
+    }
+
+    std::pair<INDEX_2, T> operator*()
+    {
+      return std::make_pair(ht.hash[bagnr][pos], ht.cont[bagnr][pos]);
     }
 
     bool operator != (int i) const
@@ -246,6 +265,18 @@ public:
     return GetNBags();
   }
 
+  Iterator begin () const
+  {
+    Iterator it(*this, 0, -1);
+    it++;
+    return it;
+  }
+
+  int end() const
+  {
+    return GetNBags();
+  }
+
   void GetData (const Iterator & it,
 		INDEX_2 & ahash, T & acont) const
   {
@@ -259,20 +290,13 @@ public:
   const T & GetData (const Iterator & it) const
   { return cont[it.BagNr()][it.Pos()]; }
 
-  ngstd::Archive & DoArchive (ngstd::Archive & ar)
+  void DoArchive (Archive & ar)
   {
     ar & hash & cont;
-    return ar;
   }    
   
 };
 
-  template <typename T>
-  inline ngstd::Archive & operator & (ngstd::Archive & archive, INDEX_2_HASHTABLE<T> & mp)
-  { return mp.DoArchive(archive);   }
-
-
-
 template <typename T> 
 inline ostream & operator<< (ostream & ost, const INDEX_2_HASHTABLE<T> & ht)
 {
@@ -390,9 +414,14 @@ public:
     int BagNr() const { return bagnr; }
     int Pos() const { return pos; }
 
-    void operator++ (int)
+    Iterator operator++ (int)
+    {
+      Iterator it(ht, bagnr, pos);
+      ++(*this);
+      return it;
+    }
+    Iterator& operator++()
     {
-      // cout << "begin Operator ++: bagnr = " << bagnr << " -  pos = " << pos << endl;
       pos++;
       while (bagnr < ht.GetNBags() && 
 	     pos == ht.GetBagSize(bagnr+1))
@@ -400,7 +429,12 @@ public:
 	  pos = 0;
 	  bagnr++;
 	}
-      // cout << "end Operator ++: bagnr = " << bagnr << " - pos = " << pos << endl;
+      return *this;
+    }
+
+    std::pair<INDEX_3, T> operator*()
+    {
+      return std::make_pair(ht.hash[bagnr][pos], ht.cont[bagnr][pos]);
     }
 
     bool operator != (int i) const
@@ -422,6 +456,18 @@ public:
     return GetNBags();
   }
 
+  Iterator begin () const
+  {
+    Iterator it(*this, 0, -1);
+    it++;
+    return it;
+  }
+
+  int end() const
+  {
+    return GetNBags();
+  }
+
   void GetData (const Iterator & it,
 		INDEX_3 & ahash, T & acont) const
   {
@@ -436,24 +482,15 @@ public:
   { return cont[it.BagNr()][it.Pos()]; }
 
 
-  ngstd::Archive & DoArchive (ngstd::Archive & ar)
+  void DoArchive (Archive & ar)
   {
     ar & hash & cont;
-    return ar;
   }    
 
 
 };
 
 
-
-  template <typename T>
-  inline ngstd::Archive & operator & (ngstd::Archive & archive, INDEX_3_HASHTABLE<T> & mp)
-  { return mp.DoArchive(archive);   }
-
-
-
-
 template <typename T> 
 inline ostream & operator<< (ostream & ost, const INDEX_3_HASHTABLE<T> & ht)
 {
@@ -496,7 +533,7 @@ class BASE_INDEX_CLOSED_HASHTABLE
 protected:
   ///
   // MoveableArray<INDEX> hash;
-  Array<INDEX> hash;
+  NgArray<INDEX> hash;
   ///
   int invalid;
 public:
@@ -572,7 +609,7 @@ class INDEX_CLOSED_HASHTABLE : public BASE_INDEX_CLOSED_HASHTABLE
 {
   ///
 // MoveableArray<T> cont;
-  Array<T> cont;
+  NgArray<T> cont;
 
 public:
   ///
@@ -669,13 +706,13 @@ class BASE_INDEX_2_CLOSED_HASHTABLE
 protected:
   ///
   // MoveableArray<INDEX_2> hash;
-  Array<INDEX_2> hash;
+  NgArray<INDEX_2> hash;
   ///
   int invalid;
   size_t mask;
 public:
   ///
-  BASE_INDEX_2_CLOSED_HASHTABLE (size_t size);
+  DLL_HEADER BASE_INDEX_2_CLOSED_HASHTABLE (size_t size);
 
   int Size() const { return hash.Size(); }
   bool UsedPos0 (int pos) const { return ! (hash[pos].I1() == invalid); }
@@ -725,16 +762,16 @@ public:
 protected:
   ///
 
-  int Position2 (const INDEX_2 & ind) const;
-  bool PositionCreate2 (const INDEX_2 & ind, int & apos);
-  void BaseSetSize (int asize);
+  DLL_HEADER int Position2 (const INDEX_2 & ind) const;
+  DLL_HEADER bool PositionCreate2 (const INDEX_2 & ind, int & apos);
+  DLL_HEADER void BaseSetSize (int asize);
 };
 
 
 template <class T>
 class INDEX_2_CLOSED_HASHTABLE : public BASE_INDEX_2_CLOSED_HASHTABLE
 {
-  Array<T> cont;
+  NgArray<T> cont;
 public:
   INDEX_2_CLOSED_HASHTABLE (size_t size)
     : BASE_INDEX_2_CLOSED_HASHTABLE(size), cont(RoundUp2(size))
@@ -759,6 +796,15 @@ public:
     int pos = Position0 (ahash);
     return (pos != -1);
   }
+
+  inline optional<T> GetIfUsed (const INDEX_2 & ahash) const
+  {
+    int pos = Position0 (ahash);
+    if (pos != -1)
+      return cont[pos];
+    else
+      return nullopt;
+  }
     
   inline void SetData0 (int pos, const INDEX_2 & ahash, const T & acont)
   {
@@ -829,7 +875,7 @@ inline ostream & operator<< (ostream & ost, const INDEX_2_CLOSED_HASHTABLE<T> &
 class BASE_INDEX_3_CLOSED_HASHTABLE
 {
 protected:
-  Array<INDEX_3> hash;
+  NgArray<INDEX_3> hash;
   int invalid;
   size_t mask;
 
@@ -938,7 +984,7 @@ template <class T>
 class INDEX_3_CLOSED_HASHTABLE : public BASE_INDEX_3_CLOSED_HASHTABLE
 {
   // MoveableArray<T,0> cont;
-  Array<T,0> cont;
+  NgArray<T,0> cont;
 
 public:
   INDEX_3_CLOSED_HASHTABLE (int size)
@@ -1377,6 +1423,10 @@ inline void SetInvalid (INDEX_2 & i2) { i2[0] = -1; }
 inline bool IsInvalid (INDEX_2 i2) { return i2[0] == -1; }
 inline size_t HashValue (INDEX_2 i2, size_t size) { return (113*size_t(i2[0])+size_t(i2[1])) % size; }
 
+inline void SetInvalid (INDEX_3 & i3) { i3[0] = -1; }
+inline bool IsInvalid (INDEX_3 i3) { return i3[0] == -1; }
+inline size_t HashValue (INDEX_3 i3, size_t size) { return (i3[0]+15*size_t(i3[1])+41*size_t(i3[2])) % size; }
+
 
   /**
      A closed hash-table.
@@ -1384,7 +1434,7 @@ inline size_t HashValue (INDEX_2 i2, size_t size) { return (113*size_t(i2[0])+si
      The array should be allocated with the double size of the expected number of entries.
   */
   template <class T_HASH, class T>
-  class ClosedHashTable
+  class NgClosedHashTable
   {
   protected:
     ///
@@ -1392,22 +1442,21 @@ inline size_t HashValue (INDEX_2 i2, size_t size) { return (113*size_t(i2[0])+si
     ///
     size_t used;
     ///
-    Array<T_HASH> hash;
+    NgArray<T_HASH> hash;
     ///
-    Array<T> cont;
+    NgArray<T> cont;
   public:
     ///
-    ClosedHashTable (size_t asize = 128)
+    NgClosedHashTable (size_t asize = 128)
       : size(asize), used(0), hash(asize), cont(asize)
     {
       for (auto & v : hash)
         SetInvalid(v);
     }
 
-    ClosedHashTable (ClosedHashTable && ht2) = default;
+    NgClosedHashTable (NgClosedHashTable && ht2) = default;
 
-      // who needs that ? 
-    ClosedHashTable (FlatArray<T_HASH> _hash, FlatArray<T> _cont)
+    NgClosedHashTable (NgFlatArray<T_HASH> _hash, NgFlatArray<T> _cont)
       : size(_hash.Size()), used(0), hash(_hash.Size(), _hash.Addr(0)), cont(_cont.Size(), _cont.Addr(0))
     {
       for (auto & v : hash)
@@ -1415,7 +1464,7 @@ inline size_t HashValue (INDEX_2 i2, size_t size) { return (113*size_t(i2[0])+si
     }
 
 
-    ClosedHashTable & operator= (ClosedHashTable && ht2) = default;
+    NgClosedHashTable & operator= (NgClosedHashTable && ht2) = default;
 
     /// 
     size_t Size() const
@@ -1433,13 +1482,6 @@ inline size_t HashValue (INDEX_2 i2, size_t size) { return (113*size_t(i2[0])+si
     size_t UsedElements () const
     {
       return used;
-      /*
-      size_t cnt = 0;
-      for (size_t i = 0; i < size; i++)
-	if (hash[i] != invalid)
-	  cnt++;
-      return cnt;
-      */
     }
 
     size_t Position (const T_HASH ind) const
@@ -1456,10 +1498,10 @@ inline size_t HashValue (INDEX_2 i2, size_t size) { return (113*size_t(i2[0])+si
 
     void DoubleSize()
     {
-      ClosedHashTable tmp(2*Size());
+      NgClosedHashTable tmp(2*Size());
       for (auto both : *this)
         tmp[both.first] = both.second;
-      *this = move(tmp);
+      *this = std::move(tmp);
     }
     
     // returns true if new position is created
@@ -1581,13 +1623,20 @@ inline size_t HashValue (INDEX_2 i2, size_t size) { return (113*size_t(i2[0])+si
           pos = nextpos;
         }
     }
+
+    void DeleteData()
+    {
+      for (auto & v : hash)
+        SetInvalid(v);
+      used = 0;
+    }
     
     class Iterator
     {
-      const ClosedHashTable & tab;
+      const NgClosedHashTable & tab;
       size_t nr;
     public:
-      Iterator (const ClosedHashTable & _tab, size_t _nr)
+      Iterator (const NgClosedHashTable & _tab, size_t _nr)
         : tab(_tab), nr(_nr)
       {
         while (nr < tab.Size() && !tab.UsedPos(nr)) nr++;
@@ -1614,7 +1663,7 @@ inline size_t HashValue (INDEX_2 i2, size_t size) { return (113*size_t(i2[0])+si
 
   template <class T_HASH, class T>  
   ostream & operator<< (ostream & ost,
-                        const ClosedHashTable<T_HASH,T> & tab)
+                        const NgClosedHashTable<T_HASH,T> & tab)
   {
     for (size_t i = 0; i < tab.Size(); i++)
       if (tab.UsedPos(i))
