@@ -27,7 +27,7 @@ namespace netgen
     oneblock = NULL;
   }
 
-  BASE_TABLE :: BASE_TABLE (const FlatArray<int> & entrysizes, int elemsize)
+  BASE_TABLE :: BASE_TABLE (const NgFlatArray<int> & entrysizes, int elemsize)
     : data(entrysizes.Size())
   {
     size_t cnt = 0;
@@ -108,8 +108,9 @@ namespace netgen
     if (line.size == line.maxsize)
       {
 	void * p = new char [(line.maxsize+5) * elsize];
-      
-	memcpy (p, line.col, line.maxsize * elsize);
+
+        if (line.maxsize*elsize)
+          memcpy (p, line.col, line.maxsize * elsize);
 	delete [] (char*)line.col;
 
 	line.col = p;
@@ -213,7 +214,7 @@ namespace netgen
 
 
   
-  ngstd::Archive & BASE_TABLE :: DoArchive (ngstd::Archive & ar, int elemsize)
+  void BASE_TABLE :: DoArchive (Archive & ar, int elemsize)
   {
     if (ar.Output())
       {
@@ -248,7 +249,6 @@ namespace netgen
             cnt += data[i].size*elemsize;
           }
       }
-    return ar;
   }    
 
   
