@@ -49,7 +49,8 @@ namespace netgen
   {
     data = NULL; height = width = 0;
     SetSize (m2.Height(), m2.Width());
-    memcpy (data, m2.data, sizeof(double) * Height() * Width());
+    if (Height()*Width())
+      memcpy (data, m2.data, sizeof(double) * (Height() * Width()));
   }
 
   DenseMatrix :: ~DenseMatrix ()
@@ -427,7 +428,7 @@ namespace netgen
         double max, hr;
       
 
-        Array<int> p(n);   // pivot-permutation
+        NgArray<int> p(n);   // pivot-permutation
         Vector hv(n);
     
       
@@ -1154,7 +1155,7 @@ namespace netgen
   }
 
 
-  void DenseMatrix :: MultElementMatrix (const Array<int> & pnum, 
+  void DenseMatrix :: MultElementMatrix (const NgArray<int> & pnum, 
                                          const Vector & hx, Vector & hy)
   {
     int i, j;
@@ -1180,7 +1181,7 @@ namespace netgen
     
   }
   
-  void DenseMatrix :: MultTransElementMatrix (const Array<int> & pnum, 
+  void DenseMatrix :: MultTransElementMatrix (const NgArray<int> & pnum, 
                                               const Vector & hx, Vector & hy)
   {
     int i, j;
@@ -1380,6 +1381,6 @@ namespace netgen
     return ost;
   }
 
-
+  
 
 }
