@@ -11,21 +11,27 @@
   Advancing front class for volume meshing
 */
 
+#include <gprim/geomobjects.hpp>
+#include <gprim/adtree.hpp>
+#include "meshtype.hpp"
+#include "geomsearch.hpp"
 
+namespace netgen
+{
 
 /// Point in advancing front
 class FrontPoint3
 {
   /// coordinates
-Point<3> p;           
+  Point<3> p;           
   /// global node index
-PointIndex globalindex;   
+  PointIndex globalindex;   
   /// number of faces connected to point 
-int nfacetopoint;    
+  int nfacetopoint;    
   /// distance to original boundary
-int frontnr;
+  int frontnr;
   /// 
-int cluster;
+  int cluster;
 public:
   ///
   FrontPoint3 ();
@@ -43,7 +49,7 @@ public:
   void AddFace ()
   { nfacetopoint++; }
 
-  ///
+  /// if last face is removed, then point is invalidated
   void RemoveFace()
   { 
     nfacetopoint--;
@@ -51,7 +57,7 @@ public:
   }
   
   ///
-  int Valid () const
+  bool Valid () const
   { return nfacetopoint >= 0; }
 
   ///
@@ -74,7 +80,7 @@ class MiniElement2d
 {
 protected:
   int np;
-  PointIndex pnum[4];
+  PointIndex pnum[4]; // can be global or local nums
   bool deleted;
 public:
   MiniElement2d ()
@@ -89,8 +95,8 @@ public:
   const PointIndex PNum (int i) const { return pnum[i-1]; }
   PointIndex & PNum (int i) { return pnum[i-1]; }
   const PointIndex PNumMod (int i) const { return pnum[(i-1)%np]; }
-
-  void Delete () { deleted = 1; pnum[0] = pnum[1] = pnum[2] = pnum[3] = PointIndex::BASE-1; }
+  auto PNums() const { return NgFlatArray<const PointIndex> (np, &pnum[0]); }
+  void Delete () { deleted = true; for (PointIndex & p : pnum) p.Invalidate(); }
   bool IsDeleted () const { return deleted; }
 };
 
@@ -120,7 +126,7 @@ private:
   int hashvalue;
   ///
   int cluster;
-
+  
 public:
   ///
   FrontFace ();
@@ -176,51 +182,51 @@ public:
 class AdFront3
 {
   ///
-  Array<FrontPoint3, PointIndex::BASE, PointIndex> points;
+  NgArray<FrontPoint3, PointIndex::BASE, PointIndex> points;
   ///
-Array<FrontFace> faces;
+  NgArray<FrontFace> faces;
   ///
-Array<PointIndex> delpointl;
-
+  NgArray<PointIndex> delpointl;
+  
   /// which points are connected to pi ?
-TABLE<int, PointIndex::BASE> * connectedpairs;
+  TABLE<int, PointIndex::BASE> * connectedpairs;
   
   /// number of total front faces;
-int nff;
+  int nff;
   /// number of quads in front
-int nff4; 
+  int nff4; 
   
   ///
-double vol;
-
+  double vol;
+  
   ///
-GeomSearch3d hashtable;
-
+  GeomSearch3d hashtable;
+  
   /// 
-int hashon;
+  int hashon;
 
   ///
-int hashcreated;
-
+  int hashcreated;
+  
   /// counter for rebuilding internal tables
-int rebuildcounter;
+  int rebuildcounter;
   /// last base element
-int lasti;
+  int lasti;
   /// minimal selection-value of baseelements
-int minval;
-  Array<int, PointIndex::BASE, PointIndex> invpindex;
-  Array<char> pingroup;
+  int minval;
+  NgArray<PointIndex, PointIndex::BASE, PointIndex> invpindex;
+  NgArray<char, PointIndex::BASE> pingroup;
   
   ///
-class Box3dTree * facetree;
+  class BoxTree<3> * facetree;
 public:
-
+  
   ///
   AdFront3 ();
   ///
   ~AdFront3 ();
   ///
-  void GetPoints (Array<Point<3> > & apoints) const;
+  void GetPoints (NgArray<Point<3> > & apoints) const;
   ///
   int GetNP() const 
   { return points.Size(); }
@@ -233,6 +239,7 @@ public:
   ///
   const MiniElement2d & GetFace (int i) const
   { return faces.Get(i).Face(); }
+  const auto & Faces() const { return faces; }
   ///
   void Print () const;
   ///
@@ -253,17 +260,20 @@ public:
 
   ///
   void GetIntersectingFaces (const Point<3> & pmin, const Point<3> & pmax, 
-			     Array<int> & ifaces) const;
+			     NgArray<int> & ifaces) const;
+
+  bool PointInsideGroup(const NgArray<PointIndex, PointIndex::BASE> &grouppindex,
+                        const NgArray<MiniElement2d>& groupfaces) const;
 
   ///
   void GetFaceBoundingBox (int i, Box3d & box) const;
 
   ///
   int GetLocals (int baseelement,
-		 Array<Point3d > & locpoints,
-                 Array<MiniElement2d> & locfaces,   // local index
-                 Array<PointIndex> & pindex,
-                 Array<INDEX> & findex,
+		 NgArray<Point3d, PointIndex::BASE> & locpoints,
+                 NgArray<MiniElement2d> & locfaces,   // local index
+                 NgArray<PointIndex, PointIndex::BASE> & pindex,
+                 NgArray<INDEX> & findex,
 		 INDEX_2_HASHTABLE<int> & connectedpairs,
                  float xh,
 		 float relh,
@@ -271,10 +281,10 @@ public:
   
   ///
   void GetGroup (int fi,
-                 Array<MeshPoint> & grouppoints,
-                 Array<MiniElement2d> & groupelements,
-                 Array<PointIndex> & pindex,
-                 Array<INDEX> & findex);
+                 NgArray<MeshPoint, PointIndex::BASE> & grouppoints,
+                 NgArray<MiniElement2d> & groupelements,
+                 NgArray<PointIndex, PointIndex::BASE> & pindex,
+                 NgArray<INDEX> & findex);
 
   ///
   void DeleteFace (INDEX fi);
@@ -299,7 +309,7 @@ public:
   bool Inside (const Point<3> & p) const;
   /// both points on same side ?
   int SameSide (const Point<3> & lp1, const Point<3> & lp2, 
-		const Array<int> * testfaces = NULL) const;
+		const NgArray<int> * testfaces = NULL) const;
 
 
   ///
@@ -314,7 +324,5 @@ private:
   void RebuildInternalTables();
 };
 
-
-
-
+} // namespace netgen
 #endif
