@@ -66,7 +66,7 @@ namespace netgen
           auto p0 = larger_trig[i];
           auto p1 = larger_trig[(i+1)%3];
           auto p2 = larger_trig[(i+2)%3];
-          auto n = Cross(p2-p1, n_trig);
+          // auto n = Cross(p2-p1, n_trig);
 
           auto v0 = (p2-p1).Normalize();
           auto v1 = (p0-p1).Normalize();
@@ -209,7 +209,7 @@ namespace netgen
         Vec<3> ab_base_norm = (b_base - a_base).Normalize();
         double a_vec_x = Dot(a_vec, ab_base_norm);
         double b_vec_x = Dot(b_vec, -ab_base_norm);
-        double ratio_parallel = (a_vec_x + b_vec_x) / ab_base;
+        // double ratio_parallel = (a_vec_x + b_vec_x) / ab_base;
 
         // Calculate surface normal at point si
         Vec<3> surface_normal = getNormal(mesh[si]);
@@ -238,7 +238,7 @@ namespace netgen
     };
     
     auto modifiedsmooth = [&](size_t nsteps) {
-        for (auto i : Range(nsteps))
+        for ([[maybe_unused]] auto i : Range(nsteps))
             for (SurfaceElementIndex sei : mesh.SurfaceElements().Range())
             {
             // assuming triangle
@@ -248,8 +248,9 @@ namespace netgen
         }
     };
 
+    /*
     auto smooth = [&] (size_t nsteps) {
-        for(auto i : Range(nsteps))
+      for([[maybe_unused]] auto i : Range(nsteps))
             for(const auto & sel : mesh.SurfaceElements())
             {
                 double min_limit = 999;
@@ -259,13 +260,14 @@ namespace netgen
                     limits[pi] = min(limits[pi], 1.4*min_limit);
             }
     };
-
+    */
+    
     // check for self-intersection within new elements (prisms/hexes)
     auto self_intersection = [&] () {
         for(SurfaceElementIndex sei : mesh.SurfaceElements().Range())
           {
             auto facei = mesh[sei].GetIndex();
-            if(facei < nfd_old && !params.surfid.Contains(facei))
+            if(facei < nfd_old && !par_surfid.Contains(facei))
                 continue;
 
             auto sel = mesh[sei];
@@ -320,7 +322,7 @@ namespace netgen
             for(auto pi : sel.PNums())
                 box.Add(mesh[pi]);
             // also add moved points to bounding box
-            if(params.surfid.Contains(sel.GetIndex()))
+            if(par_surfid.Contains(sel.GetIndex()))
                 for(auto pi : sel.PNums())
                     box.Add(mesh[pi]+limits[pi]*height*growthvectors[pi]);
             tree.Insert(box, sei);
@@ -344,7 +346,7 @@ namespace netgen
                     return false;
                 auto face = GetMappedFace(sei, -2);
                 double lam_ = 999;
-                bool is_bl_sel = params.surfid.Contains(sel.GetIndex());
+                bool is_bl_sel = par_surfid.Contains(sel.GetIndex());
 
                 if (step == 0)
                 {
@@ -446,7 +448,7 @@ namespace netgen
   Array<Segment> BuildSegments( Mesh & mesh )
   {
       Array<Segment> segments;
-      auto& topo = mesh.GetTopology();
+      // auto& topo = mesh.GetTopology();
 
       NgArray<SurfaceElementIndex> surf_els;
 
@@ -519,7 +521,7 @@ namespace netgen
         for(SurfaceElementIndex sei : myrange)
           {
             auto facei = mesh[sei].GetIndex();
-            if(facei < nfd_old && !params.surfid.Contains(facei))
+            if(facei < nfd_old && !par_surfid.Contains(facei))
               {
                 for(auto pi : mesh[sei].PNums())
                   if(mesh[pi].Type() == SURFACEPOINT)
@@ -550,13 +552,13 @@ namespace netgen
 
     // smooth tangential part of growth vectors from edges to surface elements
     RegionTimer rtsmooth(tsmooth);
-    for(auto i : Range(10))
+    for([[maybe_unused]] auto i : Range(10))
     {
         for(auto pi : points)
         {
             auto sels = p2sel[pi];
             Vec<3> new_gw = growthvectors[pi];
-            int cnt = 1;
+            // int cnt = 1;
             std::set<PointIndex> suround;
             suround.insert(pi);
             auto normal = normals[pi];
@@ -591,12 +593,13 @@ namespace netgen
   {
     static Timer timer("BoundaryLayerTool::ctor");
     RegionTimer regt(timer);
+    ProcessParameters();
 
     //for(auto & seg : mesh.LineSegments())
         //seg.edgenr = seg.epgeominfo[1].edgenr;
 
     height = 0.0;
-    for (auto h : params.heights)
+    for (auto h : par_heights)
       height += h;
 
     max_edge_nr = -1;
@@ -609,7 +612,7 @@ namespace netgen
 
     new_mat_nrs.SetSize(mesh.FaceDescriptors().Size() + 1);
     new_mat_nrs = -1;
-    for(auto [bcname, matname] : params.new_mat)
+    for(auto [bcname, matname] : par_new_mat)
       {
         mesh.SetMaterial(++ndom, matname);
         regex pattern(bcname);
@@ -621,7 +624,6 @@ namespace netgen
           }
       }
 
-    domains = params.domains;
     if(!params.outside)
       domains.Invert();
 
@@ -658,7 +660,7 @@ namespace netgen
       {
         const auto& fd = mesh.GetFaceDescriptor(i);
         string name = fd.GetBCName();
-        if(params.surfid.Contains(i))
+        if(par_surfid.Contains(i))
           {
             if(auto isIn = domains.Test(fd.DomainIn()); isIn != domains.Test(fd.DomainOut()))
               {
@@ -668,15 +670,22 @@ namespace netgen
                 FaceDescriptor new_fd(-1, isIn ? new_mat_nrs[i] : fd.DomainIn(),
                                       isIn ? fd.DomainOut() : new_mat_nrs[i], -1);
                 new_fd.SetBCProperty(new_si);
+                new_fd.SetSurfColour(fd.SurfColour());
                 mesh.AddFaceDescriptor(new_fd);
                 si_map[i] = new_si;
                 moved_surfaces.SetBit(i);
                 mesh.SetBCName(new_si-1, "mapped_" + name);
               }
+            // curving of surfaces with boundary layers will often
+            // result in pushed through elements, since we do not (yet)
+            // curvature through layers.
+            // Therefore we disable curving for these surfaces.
+            if(!params.keep_surfaceindex)
+              mesh.GetFaceDescriptor(i).SetSurfNr(-1);
           }
       }
 
-    for(auto si : params.surfid)
+    for(auto si : par_surfid)
       if(surfacefacs[si] == 0.0)
         throw Exception("Surface " + to_string(si) + " is not a boundary of the domain to be grown into!");
   }
@@ -691,20 +700,22 @@ namespace netgen
         if(face_done.Test(facei))
           continue;
         bool point_moved = false;
-        bool point_fixed = false;
+        // bool point_fixed = false;
         for(auto pi : sel.PNums())
           {
             if(growthvectors[pi].Length() > 0)
               point_moved = true;
+            /*
             else
               point_fixed = true;
+            */
           }
         if(point_moved && !moved_surfaces.Test(facei))
           {
             int new_si = mesh.GetNFD()+1;
             const auto& fd = mesh.GetFaceDescriptor(facei);
-            auto isIn = domains.Test(fd.DomainIn());
-            auto isOut = domains.Test(fd.DomainOut());
+            // auto isIn = domains.Test(fd.DomainIn());
+            // auto isOut = domains.Test(fd.DomainOut());
             int si = params.sides_keep_surfaceindex ? facei : -1;
             // domin and domout can only be set later
             FaceDescriptor new_fd(si, -1,
@@ -736,7 +747,7 @@ namespace netgen
         {
             const auto & sel = mesh[sei];
             auto facei = sel.GetIndex();
-            if(!params.surfid.Contains(facei))
+            if(!par_surfid.Contains(facei))
                 continue;
 
             auto n = surfacefacs[sel.GetIndex()] * getNormal(sel);
@@ -932,7 +943,7 @@ namespace netgen
                   else
                     continue;
 
-                  if(!params.project_boundaries.Contains(sel.GetIndex()))
+                  if(!par_project_boundaries.Contains(sel.GetIndex()))
                     continue;
                   auto& g = growthvectors[pi];
                   auto ng = n * g;
@@ -951,7 +962,7 @@ namespace netgen
           {
             int count = 0;
             for(const auto& seg2 : segments)
-              if(((seg[0] == seg2[0] && seg[1] == seg2[1]) || (seg[0] == seg2[1] && seg[1] == seg2[0])) && params.surfid.Contains(seg2.si))
+              if(((seg[0] == seg2[0] && seg[1] == seg2[1]) || (seg[0] == seg2[1] && seg[1] == seg2[0])) && par_surfid.Contains(seg2.si))
                 count++;
             if(count == 1)
               {
@@ -981,10 +992,12 @@ namespace netgen
           //   return true;
           // return false;
             auto segs = topo.GetVertexSegments(pi);
+            if(segs.Size() == 1)
+              return true;
             auto first_edgenr = mesh[segs[0]].edgenr;
             for(auto segi : segs)
-                if(mesh[segi].edgenr != first_edgenr)
-                    return true;
+              if(mesh[segi].edgenr != first_edgenr)
+                return true;
             return false;
         };
 
@@ -1086,9 +1099,9 @@ namespace netgen
       if (growthvectors[pi].Length2() != 0)
         {
           Point<3> p = mesh[pi];
-          for(auto i : Range(params.heights))
+          for(auto i : Range(par_heights))
             {
-              p += params.heights[i] * growthvectors[pi];
+              p += par_heights[i] * growthvectors[pi];
               mapto[pi].Append(mesh.AddPoint(p));
             }
         }
@@ -1101,7 +1114,7 @@ namespace netgen
           {
             // copy here since we will add segments and this would
             // invalidate a reference!
-            auto segi = segments[sei];
+            // auto segi = segments[sei];
             for(auto [sej, type] : segmap[sei])
               {
                 auto segj = segments[sej];
@@ -1139,7 +1152,7 @@ namespace netgen
                     s0.si = segj.si;
                     new_segments.Append(s0);
 
-                    for(auto i : Range(params.heights))
+                    for(auto i : Range(par_heights))
                       {
                         Element2d sel(QUAD);
                         p3 = mapto[pp2][i];
@@ -1207,7 +1220,7 @@ namespace netgen
           {
             Array<PointIndex> points(sel.PNums());
             if(surfacefacs[sel.GetIndex()] > 0) Swap(points[0], points[2]);
-            for(auto j : Range(params.heights))
+            for(auto j : Range(par_heights))
               {
                 auto eltype = points.Size() == 3 ? PRISM : HEX;
                 Element el(eltype);
@@ -1218,6 +1231,9 @@ namespace netgen
                 if(surfacefacs[sel.GetIndex()] > 0) Swap(points[0], points[2]);
                 for(auto i : Range(points))
                   el[sel.PNums().Size() + i] = points[i];
+                auto new_index = new_mat_nrs[sel.GetIndex()];
+                if(new_index == -1)
+                  throw Exception("Boundary " + ToString(sel.GetIndex()) + " with name " + mesh.GetBCName(sel.GetIndex()-1) + " extruded, but no new material specified for it!");
                 el.SetIndex(new_mat_nrs[sel.GetIndex()]);
                 mesh.AddVolumeElement(el);
               }
@@ -1312,7 +1328,7 @@ namespace netgen
                     PointIndex p4 = p1;
                     PointIndex p5 = p2;
                     PointIndex p6 = p3;
-                    for(auto i : Range(params.heights))
+                    for(auto i : Range(par_heights))
                       {
                         Element nel(PRISM);
                         nel[0] = p4; nel[1] = p5; nel[2] = p6;
@@ -1328,7 +1344,7 @@ namespace netgen
                       {
                         PointIndex p1 = moved[0];
                         PointIndex p2 = moved[1];
-                        for(auto i : Range(params.heights))
+                        for(auto i : Range(par_heights))
                           {
                             PointIndex p3 = mapto[moved[1]][i];
                             PointIndex p4 = mapto[moved[0]][i];
@@ -1350,7 +1366,7 @@ namespace netgen
                 if(moved.Size() == 1 && fixed.Size() == 1)
                   {
                     PointIndex p1 = moved[0];
-                    for(auto i : Range(params.heights))
+                    for(auto i : Range(par_heights))
                       {
                         Element nel = el;
                         PointIndex p2 = mapto[moved[0]][i];
@@ -1374,7 +1390,7 @@ namespace netgen
                       throw Exception("This case is not implemented yet! Fixed size = " + ToString(fixed.Size()));
                     PointIndex p1 = moved[0];
                     PointIndex p2 = moved[1];
-                    for(auto i : Range(params.heights))
+                    for(auto i : Range(par_heights))
                       {
                         PointIndex p3 = mapto[moved[1]][i];
                         PointIndex p4 = mapto[moved[0]][i];
@@ -1473,9 +1489,9 @@ namespace netgen
               points.Append(pi);
 
       auto p2el = mesh.CreatePoint2ElementTable(is_inner_point);
-
+      
       // smooth growth vectors to shift additional element layers to the inside and fix flipped tets
-      for(auto step : Range(10))
+      for([[maybe_unused]] auto step : Range(10))
       {
           for(auto pi : points)
           {
@@ -1501,6 +1517,107 @@ namespace netgen
       }
   }
 
+  void BoundaryLayerTool :: ProcessParameters()
+  {
+      if(int* bc = get_if<int>(&params.boundary); bc)
+        {
+          for (int i = 1; i <= mesh.GetNFD(); i++)
+            if(mesh.GetFaceDescriptor(i).BCProperty() == *bc)
+              par_surfid.Append(i);
+        }
+      else if(string* s = get_if<string>(&params.boundary); s)
+        {
+          regex pattern(*s);
+          BitArray boundaries(mesh.GetNFD()+1);
+          boundaries.Clear();
+          for(int i = 1; i<=mesh.GetNFD(); i++)
+            {
+              auto& fd = mesh.GetFaceDescriptor(i);
+              if(regex_match(fd.GetBCName(), pattern))
+                {
+                  boundaries.SetBit(i);
+                  auto dom_pattern = get_if<string>(&params.domain);
+                  // only add if adjacent to domain
+                  if(dom_pattern)
+                    {
+                      regex pattern(*dom_pattern);
+                      bool mat1_match = fd.DomainIn() > 0 && regex_match(mesh.GetMaterial(fd.DomainIn()), pattern);
+                      bool mat2_match = fd.DomainOut() > 0 && regex_match(mesh.GetMaterial(fd.DomainOut()), pattern);
+                      // if boundary is inner or outer remove from list
+                      if(mat1_match == mat2_match)
+                        boundaries.Clear(i);
+                      // if((fd.DomainIn() > 0 && regex_match(mesh.GetMaterial(fd.DomainIn()), pattern)) || (fd.DomainOut() > 0 && regex_match(self.GetMaterial(fd.DomainOut()), pattern)))
+                      // boundaries.Clear(i);
+                      // par_surfid.Append(i);
+                    }
+                  // else
+                  //   par_surfid.Append(i);
+                }
+              }
+            for(int i = 1; i<=mesh.GetNFD(); i++)
+              if(boundaries.Test(i))
+                par_surfid.Append(i);
+        }
+      else
+        {
+          auto & surfids = *get_if<std::vector<int>>(&params.boundary);
+          for(auto id : surfids)
+            par_surfid.Append(id);
+        }
+      if(string* mat = get_if<string>(&params.new_material); mat)
+          par_new_mat = { { ".*", *mat } };
+      else
+          par_new_mat = *get_if<map<string, string>>(&params.new_material);
+
+      if(params.project_boundaries.has_value())
+        {
+          auto proj_bnd = *params.project_boundaries;
+          if(string* s = get_if<string>(&proj_bnd); s)
+            {
+              regex pattern(*s);
+              for(int i = 1; i<=mesh.GetNFD(); i++)
+                if(regex_match(mesh.GetFaceDescriptor(i).GetBCName(), pattern))
+                  par_project_boundaries.Append(i);
+            }
+          else
+           {
+              for(auto id : *get_if<std::vector<int>>(&proj_bnd))
+                par_project_boundaries.Append(id);
+           }
+        }
+
+      if(double* height = get_if<double>(&params.thickness); height)
+        {
+          par_heights.Append(*height);
+        }
+      else
+        {
+          auto & heights = *get_if<std::vector<double>>(&params.thickness);
+          for(auto val : heights)
+            par_heights.Append(val);
+        }
+
+      int nr_domains = mesh.GetNDomains();
+      domains.SetSize(nr_domains + 1); // one based
+      domains.Clear();
+      if(string* pdomain = get_if<string>(&params.domain); pdomain)
+        {
+          regex pattern(*pdomain);
+          for(auto i : Range(1, nr_domains+1))
+            if(regex_match(mesh.GetMaterial(i), pattern))
+              domains.SetBit(i);
+        }
+      else if(int *idomain = get_if<int>(&params.domain); idomain)
+        {
+          domains.SetBit(*idomain);
+        }
+      else
+        {
+          for (auto i : *get_if<std::vector<int>>(&params.domain))
+            domains.SetBit(i);
+        }
+  }
+
   void BoundaryLayerTool :: Perform()
   {
       CreateNewFaceDescriptors();
