@@ -1,26 +1,54 @@
 #ifndef FILE_IMPROVE3
 #define FILE_IMPROVE3
 
-
-extern double CalcTotalBad (const Mesh::T_POINTS & points, 
-			    const Mesh::T_VOLELEMENTS & elements,
-			    const MeshingParameters & mp);
+namespace netgen
+{
 
 
 ///
 class MeshOptimize3d
 {
+  Mesh & mesh;
   const MeshingParameters & mp;
+  OPTIMIZEGOAL goal = OPT_QUALITY;
+  double min_badness = 0;
+
+  bool HasBadElement(FlatArray<ElementIndex> els);
+  bool HasIllegalElement(FlatArray<ElementIndex> els);
+  bool NeedsOptimization(FlatArray<ElementIndex> els);
+
 public:
-  MeshOptimize3d (const MeshingParameters & amp) : mp(amp) { ; }
-  void CombineImprove (Mesh & mesh, OPTIMIZEGOAL goal = OPT_QUALITY);
-  void SplitImprove (Mesh & mesh, OPTIMIZEGOAL goal = OPT_QUALITY);
-  void SwapImprove (Mesh & mesh, OPTIMIZEGOAL goal = OPT_QUALITY,
-		    const BitArray * working_elements = NULL);
-  void SwapImproveSurface (Mesh & mesh, OPTIMIZEGOAL goal = OPT_QUALITY,
-			   const BitArray * working_elements = NULL,
-			   const Array< Array<int,PointIndex::BASE>* > * idmaps = NULL);
-  void SwapImprove2 (Mesh & mesh, OPTIMIZEGOAL goal = OPT_QUALITY);
+
+  MeshOptimize3d (Mesh & m, const MeshingParameters & amp, OPTIMIZEGOAL agoal = OPT_QUALITY) :
+      mesh(m), mp(amp), goal(agoal) { ; }
+
+  void SetGoal(OPTIMIZEGOAL agoal) { goal = agoal; }
+  void SetMinBadness(double badness) { min_badness = badness; }
+
+  tuple<double, double, int> UpdateBadness();
+
+  double CombineImproveEdge (
+            Table<ElementIndex, PointIndex> & elements_of_point,
+            PointIndex pi0, PointIndex pi1,
+            FlatArray<bool, PointIndex> is_point_removed, bool check_only=false);
+
+  void CombineImprove ();
+
+  void SplitImprove ();
+  double SplitImproveEdge (Table<ElementIndex,PointIndex> & elementsonnode, NgArray<INDEX_3> &locfaces, double badmax, PointIndex pi1, PointIndex pi2, PointIndex ptmp, bool check_only=false);
+
+  void SplitImprove2 ();
+  double SplitImprove2Element (ElementIndex ei, const Table<ElementIndex, PointIndex> & elements_of_point, bool check_only);
+  
+
+  double SwapImproveEdge (const NgBitArray * working_elements, Table<ElementIndex,PointIndex> & elementsonnode, INDEX_3_HASHTABLE<int> & faces, PointIndex pi1, PointIndex pi2, bool check_only=false);
+  void SwapImprove (const NgBitArray * working_elements = NULL);
+  void SwapImproveSurface (const NgBitArray * working_elements = NULL,
+			   const NgArray< NgArray<int,PointIndex::BASE>* > * idmaps = NULL);
+  void SwapImprove2 ();
+  double SwapImprove2 (ElementIndex eli1, int face, Table<ElementIndex, PointIndex> & elementsonnode, TABLE<SurfaceElementIndex, PointIndex::BASE> & belementsonnode, bool check_only=false );
+
+  void ImproveMesh() { mesh.ImproveMesh(mp, goal); }
 
   double 
   CalcBad (const Mesh::T_POINTS & points, const Element & elem, double h)
@@ -32,10 +60,9 @@ public:
   }
 
 
-  double CalcTotalBad (const Mesh::T_POINTS & points, 
-		       const Mesh::T_VOLELEMENTS & elements)
+  double GetLegalPenalty()
   {
-    return netgen::CalcTotalBad (points, elements, mp);
+    return goal == OPT_LEGAL ? 1e15 : 1e6;
   }
 
 };
@@ -61,7 +88,7 @@ extern int WrongOrientation (const Mesh::T_POINTS & points, const Element & el);
 class MinFunctionSum : public MinFunction
 {
 protected:
-  Array<MinFunction*> functions;
+  NgArray<MinFunction*> functions;
  
 public:
   
@@ -81,12 +108,12 @@ public:
 class PointFunction1 : public MinFunction
 {
   Mesh::T_POINTS & points;
-  const Array<INDEX_3> & faces;
+  const NgArray<INDEX_3> & faces;
   const MeshingParameters & mp;
   double h;
 public:
   PointFunction1 (Mesh::T_POINTS & apoints, 
-		  const Array<INDEX_3> & afaces,
+		  const NgArray<INDEX_3> & afaces,
 		  const MeshingParameters & amp,
 		  double ah);
   
@@ -100,7 +127,7 @@ class JacobianPointFunction : public MinFunction
 {
 public:
   Mesh::T_POINTS & points;
-  const Mesh::T_VOLELEMENTS & elements;
+  const Array<Element, ElementIndex> & elements;
   TABLE<INDEX> elementsonpoint;
   PointIndex actpind;
 
@@ -109,7 +136,7 @@ public:
   
 public:
   JacobianPointFunction (Mesh::T_POINTS & apoints, 
-			 const Mesh::T_VOLELEMENTS & aelements);
+			 const Array<Element, ElementIndex> & aelements);
   virtual ~JacobianPointFunction () { ; }
   virtual void SetPointIndex (PointIndex aactpind);
   virtual double Func (const Vector & x) const;
@@ -120,6 +147,5 @@ public:
   inline void UnSetNV(void) {onplane = false;}
 };
 
-
-
+} // namespace netgen
 #endif
