@@ -76,8 +76,8 @@ Meshing3 :: ~Meshing3 ()
 
 /*
   // was war das ????
-static double CalcLocH (const Array<Point3d> & locpoints,
-			const Array<MiniElement2d> & locfaces,
+static double CalcLocH (const NgArray<Point3d> & locpoints,
+			const NgArray<MiniElement2d> & locfaces,
 			double h)
 {
   return h;
@@ -168,24 +168,29 @@ int Meshing3 :: AddConnectedPair (const INDEX_2 & apair)
 MESHING3_RESULT Meshing3 :: 
 GenerateMesh (Mesh & mesh, const MeshingParameters & mp)
 {
-  static int meshing3_timer = NgProfiler::CreateTimer ("Meshing3::GenerateMesh");
-  static int meshing3_timer_a = NgProfiler::CreateTimer ("Meshing3::GenerateMesh a");
-  static int meshing3_timer_b = NgProfiler::CreateTimer ("Meshing3::GenerateMesh b");
-  static int meshing3_timer_c = NgProfiler::CreateTimer ("Meshing3::GenerateMesh c");
-  static int meshing3_timer_d = NgProfiler::CreateTimer ("Meshing3::GenerateMesh d");
-  NgProfiler::RegionTimer reg (meshing3_timer);
+  static Timer t("Meshing3::GenerateMesh"); RegionTimer reg(t);
+  // static Timer meshing3_timer_a("Meshing3::GenerateMesh a", 2);
+  // static Timer meshing3_timer_b("Meshing3::GenerateMesh b", 2);
+  // static Timer meshing3_timer_c("Meshing3::GenerateMesh c", 1);
+  // static Timer meshing3_timer_d("Meshing3::GenerateMesh d", 2);
+  // static int meshing3_timer = NgProfiler::CreateTimer ("Meshing3::GenerateMesh");
+  // static int meshing3_timer_a = NgProfiler::CreateTimer ("Meshing3::GenerateMesh a");
+  // static int meshing3_timer_b = NgProfiler::CreateTimer ("Meshing3::GenerateMesh b");
+  // static int meshing3_timer_c = NgProfiler::CreateTimer ("Meshing3::GenerateMesh c");
+  // static int meshing3_timer_d = NgProfiler::CreateTimer ("Meshing3::GenerateMesh d");
+  // NgProfiler::RegionTimer reg (meshing3_timer);
 
 
-  Array<Point3d, PointIndex::BASE> locpoints;      // local points
-  Array<MiniElement2d> locfaces;                   // local faces
-  Array<PointIndex, PointIndex::BASE> pindex;      // mapping from local to front point numbering
-  Array<int, PointIndex::BASE> allowpoint;         // point is allowed ?
-  Array<INDEX> findex;                             // mapping from local to front face numbering
+  NgArray<Point3d, PointIndex::BASE> locpoints;      // local points
+  NgArray<MiniElement2d> locfaces;                   // local faces
+  NgArray<PointIndex, PointIndex::BASE> pindex;      // mapping from local to front point numbering
+  NgArray<int, PointIndex::BASE> allowpoint;         // point is allowed ?
+  NgArray<INDEX> findex;                             // mapping from local to front face numbering
   //INDEX_2_HASHTABLE<int> connectedpairs(100);    // connecgted pairs for prism meshing
 
-  Array<Point3d, PointIndex::BASE> plainpoints;    // points in reference coordinates
-  Array<int> delpoints, delfaces;   // points and lines to be deleted
-  Array<Element> locelements;       // new generated elements
+  NgArray<Point3d, PointIndex::BASE> plainpoints;    // points in reference coordinates
+  NgArray<int> delpoints, delfaces;   // points and lines to be deleted
+  NgArray<Element> locelements;       // new generated elements
 
   int j, oldnp, oldnf;
   int found;
@@ -206,22 +211,22 @@ GenerateMesh (Mesh & mesh, const MeshingParameters & mp)
 
   
   // for star-shaped domain meshing
-  Array<MeshPoint, PointIndex::BASE> grouppoints;      
-  Array<MiniElement2d> groupfaces;
-  Array<PointIndex, PointIndex::BASE> grouppindex;
-  Array<INDEX> groupfindex;
+  NgArray<MeshPoint, PointIndex::BASE> grouppoints;      
+  NgArray<MiniElement2d> groupfaces;
+  NgArray<PointIndex, PointIndex::BASE> grouppindex;
+  NgArray<INDEX> groupfindex;
   
   
   float minerr;
   int hasfound;
-  double tetvol;
+  [[maybe_unused]] double tetvol;
   // int giveup = 0;
 
   
-  Array<Point3d> tempnewpoints;
-  Array<MiniElement2d> tempnewfaces;
-  Array<int> tempdelfaces;
-  Array<Element> templocelements;
+  NgArray<Point3d> tempnewpoints;
+  NgArray<MiniElement2d> tempnewfaces;
+  NgArray<int> tempdelfaces;
+  NgArray<Element> templocelements;
 
 
   stat.h = mp.maxh;
@@ -269,20 +274,16 @@ GenerateMesh (Mesh & mesh, const MeshingParameters & mp)
 	}
 
       const MiniElement2d & bel = adfront->GetFace (baseelem);
-      const Point3d & p1 = adfront->GetPoint (bel[0]);
-      const Point3d & p2 = adfront->GetPoint (bel[1]);
-      const Point3d & p3 = adfront->GetPoint (bel[2]);
+      const Point<3> p1 = adfront->GetPoint (bel[0]);
+      const Point<3> p2 = adfront->GetPoint (bel[1]);
+      const Point<3> p3 = adfront->GetPoint (bel[2]);
+      
 
-      // (*testout) << endl << "base = " << bel << endl;
-
-
-      Point3d pmid = Center (p1, p2, p3);
+      Point<3> pmid = Center (p1, p2, p3);
 
       double his = (Dist (p1, p2) + Dist(p1, p3) + Dist(p2, p3)) / 3;
-      double hshould;
-
-      hshould = mesh.GetH (pmid);
-
+      double hshould = mesh.GetH (pmid);
+      
       if (adfront->GetFace (baseelem).GetNP() == 4)
 	hshould = max2 (his, hshould);
 
@@ -292,13 +293,13 @@ GenerateMesh (Mesh & mesh, const MeshingParameters & mp)
       double hinner = hmax * (1 + stat.qualclass);
       double houter = hmax * (1 + 2 * stat.qualclass);
 
-      NgProfiler::StartTimer (meshing3_timer_a);
+      // meshing3_timer_a.Start();
       stat.qualclass =
         adfront -> GetLocals (baseelem, locpoints, locfaces, 
 			      pindex, findex, connectedpairs,
 			      houter, hinner,
 			      locfacesplit);
-      NgProfiler::StopTimer (meshing3_timer_a);
+      // meshing3_timer_a.Stop();
 
       // (*testout) << "locfaces = " << endl << locfaces << endl;
 
@@ -320,9 +321,6 @@ GenerateMesh (Mesh & mesh, const MeshingParameters & mp)
 	}
 
 
-
-
-
       if (testmode)
 	{
 	  (*testout) << "baseelem = " << baseelem << " qualclass = " << stat.qualclass << endl;
@@ -357,7 +355,7 @@ GenerateMesh (Mesh & mesh, const MeshingParameters & mp)
       if (stat.qualclass >= mp.starshapeclass &&
 	  mp.baseelnp != 4)   
 	{
-	  NgProfiler::RegionTimer reg1 (meshing3_timer_b);
+	  // NgProfiler::RegionTimer reg1 (meshing3_timer_b);
 	  // star-shaped domain removing
 
 	  grouppoints.SetSize (0);
@@ -374,7 +372,8 @@ GenerateMesh (Mesh & mesh, const MeshingParameters & mp)
 	      onlytri = 0;
 	  
 	  if (onlytri && groupfaces.Size() <= 20 + 2*stat.qualclass &&
-	      FindInnerPoint (grouppoints, groupfaces, inp))
+	      FindInnerPoint (grouppoints, groupfaces, inp) &&
+              !adfront->PointInsideGroup(grouppindex, groupfaces))
 	    {
 	      (*testout) << "inner point found" << endl;
 
@@ -479,7 +478,8 @@ GenerateMesh (Mesh & mesh, const MeshingParameters & mp)
 	      (*testout) << endl;
 	    }
 
-	  NgProfiler::StartTimer (meshing3_timer_c);	  
+	  // NgProfiler::StartTimer (meshing3_timer_c);
+          // meshing3_timer_c.Start();
 
 	  found = ApplyRules (plainpoints, allowpoint, 
 			      locfaces, locfacesplit, connectedpairs,
@@ -489,12 +489,12 @@ GenerateMesh (Mesh & mesh, const MeshingParameters & mp)
 	  if (found >= 0) impossible = 0;
 	  if (found < 0) found = 0;
 
-
-	  NgProfiler::StopTimer (meshing3_timer_c);	  
+          // meshing3_timer_c.Stop();
+	  // NgProfiler::StopTimer (meshing3_timer_c);	  
 
 	  if (!found) loktestmode = 0;
 
-	  NgProfiler::RegionTimer reg2 (meshing3_timer_d);	  
+	  // NgProfiler::RegionTimer reg2 (meshing3_timer_d);	  
 	  
 	  if (loktestmode)
 	    {
@@ -733,6 +733,8 @@ enum blocktyp { BLOCKUNDEF, BLOCKINNER, BLOCKBOUND, BLOCKOUTER };
 
 void Meshing3 :: BlockFill (Mesh & mesh, double gh)
 {
+  static Timer t("Mesing3::BlockFill"); RegionTimer reg(t);
+  
   PrintMessage (3, "Block-filling called (obsolete) ");
 
   int i, j(0), i1, i2, i3, j1, j2, j3;
@@ -773,9 +775,9 @@ void Meshing3 :: BlockFill (Mesh & mesh, double gh)
   
   PrintMessage (5, "n1 = ", n1, " n2 = ", n2, " n3 = ", n3);
 
-  Array<blocktyp> inner(n);
-  Array<PointIndex> pointnr(n);
-  Array<int> frontpointnr(n);
+  NgArray<blocktyp> inner(n);
+  NgArray<PointIndex> pointnr(n);
+  NgArray<int> frontpointnr(n);
 
 
   // initialize inner to 1
@@ -1097,14 +1099,16 @@ static int TestSameSide (const Point3d & p1, const Point3d & p2)
 void Meshing3 :: BlockFillLocalH (Mesh & mesh, 
 				  const MeshingParameters & mp)
 {
+  static Timer t("Mesing3::BlockFillLocalH"); RegionTimer reg(t);
+  
   double filldist = mp.filldist;
-
-  (*testout) << "blockfill local h" << endl;
-  (*testout) << "rel filldist = " << filldist << endl;
+  
+  // (*testout) << "blockfill local h" << endl;
+  // (*testout) << "rel filldist = " << filldist << endl;
   PrintMessage (3, "blockfill local h");
 
 
-  Array<Point<3> > npoints;
+  NgArray<Point<3> > npoints;
   
   adfront -> CreateTrees();
 
@@ -1141,11 +1145,18 @@ void Meshing3 :: BlockFillLocalH (Mesh & mesh,
 
   if (mp.maxh < maxh) maxh = mp.maxh;
 
+  auto loch_ptr = mesh.LocalHFunction().Copy(bbox);
+  auto & loch = *loch_ptr;
+
   bool changed;
+  static Timer t1("loop1");
+  t1.Start();
   do 
     {
-      mesh.LocalHFunction().ClearFlags();
+      loch.ClearFlags();
 
+      static Timer tbox("adfront-bbox");
+      tbox.Start();
       for (int i = 1; i <= adfront->GetNF(); i++)
 	{
 	  const MiniElement2d & el = adfront->GetFace(i);
@@ -1158,26 +1169,28 @@ void Meshing3 :: BlockFillLocalH (Mesh & mesh,
 	  double filld = filldist * bbox.Diam();
 	  bbox.Increase (filld);
       
-      	  mesh.LocalHFunction().CutBoundary (bbox); // .PMin(), bbox.PMax());
+      	  loch.CutBoundary (bbox); // .PMin(), bbox.PMax());
 	}
+      tbox.Stop();
 
       //      locadfront = adfront;
-      mesh.LocalHFunction().FindInnerBoxes (adfront, NULL);
+      loch.FindInnerBoxes (adfront, NULL);
 
       npoints.SetSize(0);
-      mesh.LocalHFunction().GetInnerPoints (npoints);
+      loch.GetInnerPoints (npoints);
 
       changed = false;
       for (int i = 1; i <= npoints.Size(); i++)
 	{
-	  if (mesh.LocalHFunction().GetH(npoints.Get(i)) > 1.5 * maxh)
+	  if (loch.GetH(npoints.Get(i)) > 1.5 * maxh)
 	    {
-	      mesh.LocalHFunction().SetH (npoints.Get(i), maxh);
+	      loch.SetH (npoints.Get(i), maxh);
 	      changed = true;
 	    }
 	}
     }
   while (changed);
+  t1.Stop();
 
   if (debugparam.slowchecks)
     (*testout) << "Blockfill with points: " << endl;
@@ -1205,6 +1218,8 @@ void Meshing3 :: BlockFillLocalH (Mesh & mesh,
 
   // find outer points
   
+  static Timer tloch2("build loch2");
+  tloch2.Start();
   loch2.ClearFlags();
 
   for (int i = 1; i <= adfront->GetNF(); i++)
@@ -1242,6 +1257,7 @@ void Meshing3 :: BlockFillLocalH (Mesh & mesh,
       // loch2.CutBoundary (pmin, pmax);
       loch2.CutBoundary (Box<3> (pmin, pmax)); // pmin, pmax);
     }
+  tloch2.Stop();
 
   // locadfront = adfront;
   loch2.FindInnerBoxes (adfront, NULL);
