@@ -24,52 +24,6 @@ public:
 };
 
 
-#ifdef NG_MPI4PY
-#include <mpi4py.h>
-
-struct mpi4py_comm {
-  mpi4py_comm() = default;
-  mpi4py_comm(MPI_Comm value) : value(value) {}
-  operator MPI_Comm () { return value; }
-
-  MPI_Comm value;
-};
-
-namespace pybind11 { namespace detail {
-  template <> struct type_caster<mpi4py_comm> {
-    public:
-    PYBIND11_TYPE_CASTER(mpi4py_comm, _("mpi4py_comm"));
-
-      // Python -> C++
-      bool load(handle src, bool) {
-        PyObject *py_src = src.ptr();
-        // Check that we have been passed an mpi4py communicator
-        if (PyObject_TypeCheck(py_src, &PyMPIComm_Type)) {
-          // Convert to regular MPI communicator
-          value.value = *PyMPIComm_Get(py_src);
-        } else {
-          return false;
-        }
-
-        return !PyErr_Occurred();
-      }
-
-      // C++ -> Python
-      static handle cast(mpi4py_comm src,
-                         return_value_policy /* policy */,
-                         handle /* parent */)
-      {
-        // Create an mpi4py handle
-        return PyMPIComm_New(src.value);
-      }
-  };
-}} // namespace pybind11::detail
-
-#endif // NG_MPI4PY
-
-
-
-
 
 using namespace netgen;
 
@@ -101,9 +55,6 @@ static Transformation<3> global_trafo(Vec<3> (0,0,0));
 
 DLL_HEADER void ExportNetgenMeshing(py::module &m) 
 {
-#ifdef NG_MPI4PY
-  import_mpi4py();
-#endif // NG_MPI4PY
   py::register_exception<NgException>(m, "NgException");
   m.attr("_netgen_executable_started") = py::cast(netgen::netgen_executable_started);
   string script;
@@ -131,51 +82,12 @@ DLL_HEADER void ExportNetgenMeshing(py::module &m)
 
   py::implicitly_convertible<int, Identifications::ID_TYPE>();
 
-  py::class_<NgMPI_Comm> (m, "MPI_Comm")
-#ifdef NG_MPI4PY
-    .def(py::init([] (mpi4py_comm comm)
-                  {
-                    return NgMPI_Comm(comm);
-                  }))
-    .def_property_readonly ("mpi4py", [] (NgMPI_Comm comm) { return mpi4py_comm(comm); })
-#endif // NG_MPI4PY
-    .def_property_readonly ("rank", &NgMPI_Comm::Rank)
-    .def_property_readonly ("size", &NgMPI_Comm::Size)
-    .def("Barrier", &NgMPI_Comm::Barrier)
-    
-#ifdef PARALLEL
-    .def("WTime", [](NgMPI_Comm  & c) { return MPI_Wtime(); })
-#else
-    .def("WTime", [](NgMPI_Comm  & c) { return -1.0; })
-#endif
-    .def("Sum", [](NgMPI_Comm  & c, double x) { return c.AllReduce(x, MPI_SUM); })
-    .def("Min", [](NgMPI_Comm  & c, double x) { return c.AllReduce(x, MPI_MIN); })
-    .def("Max", [](NgMPI_Comm  & c, double x) { return c.AllReduce(x, MPI_MAX); })
-    .def("Sum", [](NgMPI_Comm  & c, int x) { return c.AllReduce(x, MPI_SUM); })
-    .def("Min", [](NgMPI_Comm  & c, int x) { return c.AllReduce(x, MPI_MIN); })
-    .def("Max", [](NgMPI_Comm  & c, int x) { return c.AllReduce(x, MPI_MAX); })
-    .def("Sum", [](NgMPI_Comm  & c, size_t x) { return c.AllReduce(x, MPI_SUM); })
-    .def("Min", [](NgMPI_Comm  & c, size_t x) { return c.AllReduce(x, MPI_MIN); })
-    .def("Max", [](NgMPI_Comm  & c, size_t x) { return c.AllReduce(x, MPI_MAX); })
-    .def("SubComm", [](NgMPI_Comm & c, std::vector<int> proc_list) {
-        Array<int> procs(proc_list.size());
-        for (int i = 0; i < procs.Size(); i++)
-          { procs[i] = proc_list[i]; }
-        if (!procs.Contains(c.Rank()))
-          { throw Exception("rank "+ToString(c.Rank())+" not in subcomm"); }
-	return c.SubCommunicator(procs);
-      }, py::arg("procs"));
-  ;
-
-
-#ifdef NG_MPI4PY
-  py::implicitly_convertible<mpi4py_comm, NgMPI_Comm>();
-#endif // NG_MPI4PY
-
   
   py::class_<NGDummyArgument>(m, "NGDummyArgument")
     .def("__bool__", []( NGDummyArgument &self ) { return false; } )
     ;
+
+  py::class_<LocalH, shared_ptr<LocalH>>(m, "LocalH");
   
   py::class_<Point<2>> (m, "Point2d")
     .def(py::init<double,double>())
@@ -267,6 +179,34 @@ DLL_HEADER void ExportNetgenMeshing(py::module &m)
 
   py::implicitly_convertible<py::tuple, Vec<3>>();
 
+  py::class_<Mat<3,3>>(m, "Mat33")
+    .def(py::init([](py::tuple m)
+    {
+      if(m.size() != 9)
+        throw std::length_error("Invalid dimension of input array!");
+      Mat<3,3> mat;
+      for(int i = 0; i < 3; i++)
+        for(int j = 0; j < 3; j++)
+          mat(i,j) = m[i*3+j].cast<double>();
+      return mat;
+    }))
+    .def("__getitem__", [](Mat<3,3>& mat, py::tuple index)
+    {
+      if(index.size() != 2)
+        throw std::length_error("Invalid dimension of input array!");
+      return mat(index[0].cast<int>(), index[1].cast<int>());
+    })
+    .def("__setitem__", [](Mat<3,3>& mat, py::tuple index, double val)
+    {
+      if(index.size() != 2)
+        throw std::length_error("Invalid dimension of input array!");
+      mat(index[0].cast<int>(), index[1].cast<int>()) = val;
+    })
+    .def("__str__", &ToString<Mat<3,3>>)
+    ;
+
+  py::implicitly_convertible<py::tuple, Mat<3,3>>();
+
   m.def ("Vec", FunctionPointer
            ([] (double x, double y, double z) { return global_trafo(Vec<3>(x,y,z)); }));
   m.def("Vec", [](py::array_t<double> np_array)
@@ -291,6 +231,11 @@ DLL_HEADER void ExportNetgenMeshing(py::module &m)
     .def("__mul__", [](Transformation<3> a, Transformation<3> b)->Transformation<3>
          { Transformation<3> res; res.Combine(a,b); return res; })
     .def("__call__", [] (Transformation<3> trafo, Point<3> p) { return trafo(p); })
+    .def_property("mat", &Transformation<3>::GetMatrix,
+                  [](Transformation<3>& self, const Mat<3,3>& mat)
+                  {
+                    self.GetMatrix() = mat;
+                  })
     ;
 
   m.def ("GetTransformation", [] () { return global_trafo; });
@@ -466,7 +411,11 @@ DLL_HEADER void ExportNetgenMeshing(py::module &m)
         py::detail::field_descriptor {
           "refine", data_layout["refine"], sizeof(bool),
           py::format_descriptor<bool>::format(),
-          py::detail::npy_format_descriptor<bool>::dtype() }            
+          py::detail::npy_format_descriptor<bool>::dtype() },            
+        py::detail::field_descriptor {
+          "curved", data_layout["curved"], sizeof(bool),
+          py::format_descriptor<bool>::format(),
+          py::detail::npy_format_descriptor<bool>::dtype()}            
       });
   }
 
@@ -557,6 +506,10 @@ DLL_HEADER void ExportNetgenMeshing(py::module &m)
         py::detail::field_descriptor {
           "refine", data_layout["refine"], sizeof(bool),
           py::format_descriptor<bool>::format(),
+          py::detail::npy_format_descriptor<bool>::dtype() },
+        py::detail::field_descriptor {
+          "curved", data_layout["curved"], sizeof(bool),
+          py::format_descriptor<bool>::format(),
           py::detail::npy_format_descriptor<bool>::dtype() }
       });
   }
@@ -569,10 +522,8 @@ DLL_HEADER void ExportNetgenMeshing(py::module &m)
                     for (int i = 0; i < 2; i++)
                       (*newel)[i] = py::extract<PointIndex>(vertices[i])();
                     newel -> si = index;
-                    newel -> edgenr = edgenr;
                     newel -> epgeominfo[0].edgenr = edgenr;
                     newel -> epgeominfo[1].edgenr = edgenr;
-                    // needed for codim2 in 3d
                     newel -> edgenr = index;
                     for(auto i : Range(len(trignums)))
                       newel->geominfo[i].trignum = py::cast<int>(trignums[i]);
@@ -615,14 +566,24 @@ DLL_HEADER void ExportNetgenMeshing(py::module &m)
                                      li.append (py::cast(self.surfnr2));
                                      return li;
                                    }))
-    .def_property_readonly("index", FunctionPointer([](const Segment &self) -> size_t
-		  {
-		    return self.si;
-		  }))
-    .def_property_readonly("edgenr", FunctionPointer([](const Segment & self) -> size_t
-						     {
-						       return self.edgenr;
-						     }))
+    .def_property("index",
+                  [](const Segment &self) -> size_t
+                  {
+                    return self.si;
+                  },
+                  [](Segment& self, int index)
+                  {
+                    self.si = index;
+                  })
+    .def_property("edgenr",
+                  [](const Segment & self) -> size_t
+                  {
+                    return self.edgenr;
+                  },
+                  [](Segment& self, int edgenr)
+                  {
+                    self.edgenr = edgenr;
+                  })
     .def_property("singular",
                   [](const Segment & seg) { return seg.singedge_left; },
                   [](Segment & seg, double sing) { seg.singedge_left = sing; seg.singedge_right=sing; })
@@ -1005,6 +966,24 @@ DLL_HEADER void ExportNetgenMeshing(py::module &m)
       return FlatArray<int>(self.mlparentsurfaceelement.Size(),
                             &self.mlparentsurfaceelement[0]);
     }, py::keep_alive<0,1>())
+    .def_property_readonly("macromesh", [](Mesh & self) {
+      auto coarsemesh = make_shared<Mesh>();
+      *coarsemesh = *self.coarsemesh;
+      return coarsemesh;
+    }, "mesh before hp-refinement")
+    .def("MacroElementNr", [](Mesh & self, int elnr, optional<int> dim) {
+      // cout << "hpels = " << self.hpelements->Size() << endl;
+      // return self[ElementIndex(elnr)].GetHpElnr();
+      if (!dim) dim = self.GetDimension();
+      switch (*dim)
+        {
+        case 2:
+          return (*self.hpelements)[self[SurfaceElementIndex(elnr)].GetHpElnr()].coarse_elnr;
+        case 3:
+          return (*self.hpelements)[self[ElementIndex(elnr)].GetHpElnr()].coarse_elnr;
+        }
+      throw Exception ("MacroElementNr not implemented for dim");
+    }, py::arg("elnr"), py::arg("dim")=nullopt, "number of macro element of element number elnr")
     .def("FaceDescriptor", static_cast<FaceDescriptor&(Mesh::*)(int)> (&Mesh::GetFaceDescriptor),
          py::return_value_policy::reference)
     .def("GetNFaceDescriptors", &Mesh::GetNFD)
@@ -1049,10 +1028,22 @@ DLL_HEADER void ExportNetgenMeshing(py::module &m)
             return self.AddSurfaceElement (el);
           })
 
-    .def ("Add", [](Mesh & self, const Segment & el)
+    .def ("Add", [](Mesh & self, const Segment & el, bool project_geominfo)
           {
+            if (project_geominfo)
+              {
+                auto &p1 = self[el[0]];
+                auto &p2 = self[el[1]];
+                auto geo = self.GetGeometry();
+                geo->ProjectPointEdge
+                  (0,0,p1,
+                   const_cast<EdgePointGeomInfo*>(&el.epgeominfo[0]));
+                geo->ProjectPointEdge
+                  (0,0,p2,
+                   const_cast<EdgePointGeomInfo*>(&el.epgeominfo[1]));
+              }
             return self.AddSegment (el);
-          })
+          }, py::arg("el"), py::arg("project_geominfo")=false)
           
     .def ("Add", [](Mesh & self, const Element0d & el)
           {
@@ -1104,7 +1095,8 @@ DLL_HEADER void ExportNetgenMeshing(py::module &m)
                   ptr += 3;
                 }
           })
-    .def ("AddElements", [](Mesh & self, int dim, int index, py::buffer b1, int base)
+    .def ("AddElements", [](Mesh & self, int dim, int index, py::buffer b1, int base,
+                            bool project_geometry)
           {
             static Timer timer("Mesh::AddElements");
             static Timer timercast("Mesh::AddElements casting");
@@ -1156,6 +1148,21 @@ DLL_HEADER void ExportNetgenMeshing(py::module &m)
                     for (int j = 0; j < np; j++)
                       el[j] = ptr[j]+PointIndex::BASE-base;
                     el.SetIndex(index);
+                    if(project_geometry)
+                      {
+                        // find some point in the mid of trig/quad for
+                        // quick + stable uv-projection of all points
+                        auto startp = Center(self[el[0]], self[el[1]], self[el[2]]);
+                        PointGeomInfo gi = self.GetGeometry()->ProjectPoint(index,
+                                                                            startp);
+                        for(auto i : Range(np))
+                          {
+                            el.GeomInfo()[i] = gi;
+                            self.GetGeometry()->ProjectPointGI(index,
+                                                               self[el[i]],
+                                                               el.GeomInfo()[i]);
+                          }
+                      }
                     self.AddSurfaceElement (el);
                     ptr += info.strides[0]/sizeof(int);
                   }
@@ -1187,7 +1194,8 @@ DLL_HEADER void ExportNetgenMeshing(py::module &m)
                   }
               }
             
-          }, py::arg("dim"), py::arg("index"), py::arg("data"), py::arg("base")=0)
+          }, py::arg("dim"), py::arg("index"), py::arg("data"), py::arg("base")=0,
+          py::arg("project_geometry")=false)
     
     .def ("DeleteSurfaceElement",
           [](Mesh & self, SurfaceElementIndex i)
@@ -1250,6 +1258,8 @@ DLL_HEADER void ExportNetgenMeshing(py::module &m)
 
     .def ("GetCD3Name", &Mesh::GetCD3Name)
     .def ("SetCD3Name", &Mesh::SetCD3Name)
+    .def ("SplitFacesByAdjacentDomains", &Mesh::SplitFacesByAdjacentDomains)
+    .def ("GetSubMesh", &Mesh::GetSubMesh, py::arg("domains")="", py::arg("faces")="")
     .def("GetIdentifications", [](Mesh & self) -> py::list
          {
            py::list points;
@@ -1274,7 +1284,8 @@ DLL_HEADER void ExportNetgenMeshing(py::module &m)
            py::arg("identnr"),
            py::arg("type")=Identifications::PERIODIC)
     .def("IdentifyPeriodicBoundaries", &Mesh::IdentifyPeriodicBoundaries,
-         py::arg("face1"), py::arg("face2"), py::arg("mapping"), py::arg("point_tolerance") = -1.)
+         py::arg("identification_name"), py::arg("face1"), py::arg("mapping"),
+py::arg("point_tolerance") = -1.)
     .def("GetNrIdentifications", [](Mesh& self)
                                  {
                                    return self.GetIdentifications().GetMaxNr();
@@ -1293,15 +1304,12 @@ DLL_HEADER void ExportNetgenMeshing(py::module &m)
            {
              MeshingParameters mp;
              if(pars) mp = *pars;
-             {
-               py::gil_scoped_acquire acquire;
-               CreateMPfromKwargs(mp, kwargs);
-             }
+             CreateMPfromKwargs(mp, kwargs);
+             py::gil_scoped_release gil_release;
              MeshVolume (mp, self);
              OptimizeVolume (mp, self);
            }, py::arg("mp")=nullptr,
-          meshingparameter_description.c_str(),
-          py::call_guard<py::gil_scoped_release>())
+          meshingparameter_description.c_str())
 
     .def ("OptimizeVolumeMesh", [](Mesh & self, MeshingParameters* pars)
           {
@@ -1310,7 +1318,11 @@ DLL_HEADER void ExportNetgenMeshing(py::module &m)
             else mp.optsteps3d = 5;
             OptimizeVolume (mp, self);
           }, py::arg("mp"), py::call_guard<py::gil_scoped_release>())
-
+    .def("SetLocalH",[](Mesh& self, shared_ptr<LocalH> localh, int layer)
+         {
+           self.SetLocalH(localh, layer);
+         }, py::arg("localh"), py::arg("layer")=1)
+    .def("GetLocalH", &Mesh::GetLocalH)
     .def ("OptimizeMesh2d", [](Mesh & self, MeshingParameters* pars, int faceindex)
           {
             self.CalcLocalH(0.5);
@@ -1350,7 +1362,23 @@ DLL_HEADER void ExportNetgenMeshing(py::module &m)
            }), py::arg("adaptive")=false, py::call_guard<py::gil_scoped_release>())
     
     .def("ZRefine", &Mesh::ZRefine)
-
+    .def("Split2Tets", &Mesh::Split2Tets)
+    .def ("SplitAlfeld", FunctionPointer
+          ([](Mesh & self)
+           {
+            NgLock meshlock (self.MajorMutex(), true);
+            Refinement & ref = const_cast<Refinement&> (self.GetGeometry()->GetRefinement());
+            ::netgen::HPRefinement (self, &ref, SPLIT_ALFELD, 1, 0.5, true, true);
+           }
+           ), py::call_guard<py::gil_scoped_release>())
+    .def ("SplitPowellSabin", FunctionPointer
+          ([](Mesh & self)
+           {
+            NgLock meshlock (self.MajorMutex(), true);
+            Refinement & ref = const_cast<Refinement&> (self.GetGeometry()->GetRefinement());
+            ::netgen::HPRefinement (self, &ref, SPLIT_POWELL, 1, 0.5, true, true);
+           }
+           ), py::call_guard<py::gil_scoped_release>())
     .def ("SecondOrder", [](Mesh & self)
           {
             self.GetGeometry()->GetRefinement().MakeSecondOrder(self);
@@ -1441,134 +1469,32 @@ DLL_HEADER void ExportNetgenMeshing(py::module &m)
     .def ("BuildSearchTree", &Mesh::BuildElementSearchTree,py::call_guard<py::gil_scoped_release>())
 
     .def ("BoundaryLayer2", GenerateBoundaryLayer2, py::arg("domain"), py::arg("thicknesses"), py::arg("make_new_domain")=true, py::arg("boundaries")=Array<int>{})
-    .def ("BoundaryLayer", [](Mesh & self, variant<string, int> boundary,
-                              variant<double, py::list> thickness,
+    .def ("BoundaryLayer", [](Mesh & self, variant<string, int, std::vector<int>> boundary,
+                              variant<double, std::vector<double>> thickness,
                               variant<string, map<string, string>> material,
-                              variant<string, int> domain, bool outside,
-                              optional<string> project_boundaries,
+                              variant<string, int, std::vector<int>> domain, bool outside,
+                              optional<variant<string, std::vector<int>>> project_boundaries,
                               bool grow_edges, bool limit_growth_vectors,
-                              bool sides_keep_surfaceindex)
+                              bool sides_keep_surfaceindex,
+                              bool keep_surfaceindex)
            {
              BoundaryLayerParameters blp;
-             BitArray boundaries(self.GetNFD()+1);
-             boundaries.Clear();
-             if(int* bc = get_if<int>(&boundary); bc)
-               {
-                 for (int i = 1; i <= self.GetNFD(); i++)
-                   if(self.GetFaceDescriptor(i).BCProperty() == *bc)
-                     boundaries.SetBit(i);
-               }
-             else
-               {
-                 regex pattern(*get_if<string>(&boundary));
-                 for(int i = 1; i<=self.GetNFD(); i++)
-                   {
-                     auto& fd = self.GetFaceDescriptor(i);
-                     if(regex_match(fd.GetBCName(), pattern))
-                       {
-                         boundaries.SetBit(i);
-                         auto dom_pattern = get_if<string>(&domain);
-                         // only add if adjacent to domain
-                         if(dom_pattern)
-                           {
-                             regex pattern(*dom_pattern);
-                             bool mat1_match = fd.DomainIn() > 0 && regex_match(self.GetMaterial(fd.DomainIn()), pattern);
-                             bool mat2_match = fd.DomainOut() > 0 && regex_match(self.GetMaterial(fd.DomainOut()), pattern);
-                             // if boundary is inner or outer remove from list
-                             if(mat1_match == mat2_match)
-                               boundaries.Clear(i);
-                             // if((fd.DomainIn() > 0 && regex_match(self.GetMaterial(fd.DomainIn()), pattern)) || (fd.DomainOut() > 0 && regex_match(self.GetMaterial(fd.DomainOut()), pattern)))
-                             // boundaries.Clear(i);
-                             // blp.surfid.Append(i);
-                           }
-                         // else
-                         //   blp.surfid.Append(i);
-                       }
-                     }
-               }
-             for(int i = 1; i<=self.GetNFD(); i++)
-               if(boundaries.Test(i))
-                 blp.surfid.Append(i);
-             if(string* mat = get_if<string>(&material); mat)
-                 blp.new_mat = { { ".*", *mat } };
-             else
-                 blp.new_mat = *get_if<map<string, string>>(&material);
-
-             if(project_boundaries.has_value())
-               {
-                 regex pattern(*project_boundaries);
-                 for(int i = 1; i<=self.GetNFD(); i++)
-                   if(regex_match(self.GetFaceDescriptor(i).GetBCName(), pattern))
-                     blp.project_boundaries.Append(i);
-               }
-
-             if(double* pthickness = get_if<double>(&thickness); pthickness)
-               {
-                 blp.heights.Append(*pthickness);
-               }
-             else
-               {
-                 auto thicknesses = *get_if<py::list>(&thickness);
-                 for(auto val : thicknesses)
-                   blp.heights.Append(val.cast<double>());
-               }
-
-             int nr_domains = self.GetNDomains();
-             blp.domains.SetSize(nr_domains + 1); // one based
-             blp.domains.Clear();
-             if(string* pdomain = get_if<string>(&domain); pdomain)
-               {
-                 regex pattern(*pdomain);
-                 for(auto i : Range(1, nr_domains+1))
-                   if(regex_match(self.GetMaterial(i), pattern))
-                     blp.domains.SetBit(i);
-               }
-             else
-               {
-                 auto idomain = *get_if<int>(&domain);
-                 blp.domains.SetBit(idomain);
-               }
-
+             blp.boundary = boundary;
+             blp.thickness = thickness;
+             blp.new_material = material;
+             blp.domain = domain;
              blp.outside = outside;
+             blp.project_boundaries = project_boundaries;
              blp.grow_edges = grow_edges;
              blp.limit_growth_vectors = limit_growth_vectors;
              blp.sides_keep_surfaceindex = sides_keep_surfaceindex;
-
+             blp.keep_surfaceindex = keep_surfaceindex;
              GenerateBoundaryLayer (self, blp);
              self.UpdateTopology();
            }, py::arg("boundary"), py::arg("thickness"), py::arg("material"),
           py::arg("domains") = ".*", py::arg("outside") = false,
           py::arg("project_boundaries")=nullopt, py::arg("grow_edges")=true, py::arg("limit_growth_vectors") = true, py::arg("sides_keep_surfaceindex")=false,
-          R"delimiter(
-Add boundary layer to mesh.
-
-Parameters
-----------
-
-boundary : string or int
-  Boundary name or number.
-
-thickness : float or List[float]
-  Thickness of boundary layer(s).
-
-material : str or List[str]
-  Material name of boundary layer(s).
-
-domain : str or int
-  Regexp for domain boundarylayer is going into.
-
-outside : bool = False
-  If true add the layer on the outside
-
-grow_edges : bool = False
-  Grow boundary layer over edges.
-
-project_boundaries : Optional[str] = None
-  Project boundarylayer to these boundaries if they meet them. Set
-  to boundaries that meet boundarylayer at a non-orthogonal edge and
-  layer-ending should be projected to that boundary.
-
-)delimiter")
+          py::arg("keep_surfaceindex")=false, "Add boundary layer to mesh. see help(BoundaryLayerParameters) for details.")
 
     .def_static ("EnableTableClass", [] (string name, bool set)
           {
@@ -1794,6 +1720,95 @@ project_boundaries : Optional[str] = None
 
   m.attr("debugparam") = py::cast(&debugparam);
 
+  py::class_<BoundaryLayerParameters>(m, "BoundaryLayerParameters")
+    .def(py::init([]( 
+        std::variant<string, int, std::vector<int>> boundary,
+        std::variant<double, std::vector<double>> thickness,
+        std::variant<string, std::map<string, string>> new_material,
+        std::variant<string, int, std::vector<int>> domain,
+        bool outside,
+        std::optional<std::variant<string, std::vector<int>>> project_boundaries,
+        bool grow_edges,
+        bool limit_growth_vectors,
+        bool sides_keep_surfaceindex,
+        bool keep_surfaceindex,
+        double limit_safety)
+        {
+          BoundaryLayerParameters blp;
+          blp.boundary = boundary;
+          blp.thickness = thickness;
+          blp.new_material = new_material;
+          blp.domain = domain;
+          blp.outside = outside;
+          blp.project_boundaries = project_boundaries;
+          blp.grow_edges = grow_edges;
+          blp.limit_growth_vectors = limit_growth_vectors;
+          blp.sides_keep_surfaceindex = sides_keep_surfaceindex;
+          blp.keep_surfaceindex = keep_surfaceindex;
+          blp.limit_safety = limit_safety;
+          return blp;
+        }),
+           py::arg("boundary"), py::arg("thickness"), py::arg("new_material"),
+           py::arg("domain") = ".*", py::arg("outside") = false,
+           py::arg("project_boundaries")=nullopt, py::arg("grow_edges")=true,
+           py::arg("limit_growth_vectors") = true, py::arg("sides_keep_surfaceindex")=false,
+           py::arg("keep_surfaceindex")=false, py::arg("limit_safety")=0.3,
+           R"delimiter(
+Add boundary layer to mesh.
+
+Parameters
+----------
+
+boundary : string or int
+  Boundary name or number.
+
+thickness : float or List[float]
+  Thickness of boundary layer(s).
+
+material : str or List[str]
+  Material name of boundary layer(s).
+
+domain : str or int
+  Regexp for domain boundarylayer is going into.
+
+outside : bool = False
+  If true add the layer on the outside
+
+grow_edges : bool = False
+  Grow boundary layer over edges.
+
+project_boundaries : Optional[str] = None
+  Project boundarylayer to these boundaries if they meet them. Set
+  to boundaries that meet boundarylayer at a non-orthogonal edge and
+  layer-ending should be projected to that boundary.
+
+)delimiter")
+    .def(py::init([]( const py::dict & d ) {
+      try {
+        // Call other constructor with named arguments by unpacking the dictionary
+        py::object cls = py::type::of<BoundaryLayerParameters>();
+        return cls(**d).cast<BoundaryLayerParameters>();
+      }
+      catch (py::error_already_set & e) {
+        cerr << "Error creating BoundaryLayerParameters from dict:" << endl;
+        cerr << e.what() << endl;
+        throw;
+      }
+    }))
+    .def_readwrite("boundary", &BoundaryLayerParameters::boundary)
+    .def_readwrite("thickness", &BoundaryLayerParameters::thickness)
+    .def_readwrite("new_material", &BoundaryLayerParameters::new_material)
+    .def_readwrite("domain", &BoundaryLayerParameters::domain)
+    .def_readwrite("outside", &BoundaryLayerParameters::outside)
+    .def_readwrite("project_boundaries", &BoundaryLayerParameters::project_boundaries)
+    .def_readwrite("grow_edges", &BoundaryLayerParameters::grow_edges)
+    .def_readwrite("limit_growth_vectors", &BoundaryLayerParameters::limit_growth_vectors)
+    .def_readwrite("sides_keep_surfaceindex", &BoundaryLayerParameters::sides_keep_surfaceindex)
+    .def_readwrite("keep_surfaceindex", &BoundaryLayerParameters::keep_surfaceindex)
+    .def_readwrite("limit_safety", &BoundaryLayerParameters::limit_safety)
+    ;
+  py::implicitly_convertible<py::dict, BoundaryLayerParameters>();
+
 #ifdef NG_CGNS
   m.def("ReadCGNSFile", &ReadCGNSFile, py::arg("filename"), py::arg("base")=1, "Read mesh and solution vectors from CGNS file");
   m.def("WriteCGNSFile", &WriteCGNSFile, py::arg("mesh"), py::arg("filename"), py::arg("names"), py::arg("values"), py::arg("locations"),
