@@ -8,7 +8,7 @@ extern double minother;
 extern double minwithoutother;
 
 
-  static double CalcElementBadness (const Array<Point3d, PointIndex::BASE> & points,
+  static double CalcElementBadness (const NgArray<Point3d, PointIndex::BASE> & points,
                                     const Element & elem)
 {
   double vol, l, l4, l5, l6;
@@ -49,13 +49,13 @@ extern double minwithoutother;
 
 int Meshing3 :: ApplyRules 
 (
- Array<Point3d, PointIndex::BASE> & lpoints,     // in: local points, out: old+new local points
- Array<int, PointIndex::BASE> & allowpoint,      // in: 2 .. it is allowed to use pointi, 1..will be allowed later, 0..no means
- Array<MiniElement2d> & lfaces,    // in: local faces, out: old+new local faces
+ NgArray<Point3d, PointIndex::BASE> & lpoints,     // in: local points, out: old+new local points
+ NgArray<int, PointIndex::BASE> & allowpoint,      // in: 2 .. it is allowed to use pointi, 1..will be allowed later, 0..no means
+ NgArray<MiniElement2d> & lfaces,    // in: local faces, out: old+new local faces
  INDEX lfacesplit,	       // for local faces in outer radius
  INDEX_2_HASHTABLE<int> & connectedpairs,  // connected pairs for prism-meshing
- Array<Element> & elements,    // out: new elements
- Array<INDEX> & delfaces,      // out: face indices of faces to delete
+ NgArray<Element> & elements,    // out: new elements
+ NgArray<INDEX> & delfaces,      // out: face indices of faces to delete
  int tolerance,                // quality class: 1 best 
  double sloppy,                // quality strength
  int rotind1,                  // how to rotate base element
@@ -63,8 +63,11 @@ int Meshing3 :: ApplyRules
  )
 
 {
-  NgProfiler::RegionTimer regtot(97);
+  // static Timer t("ruler3 - all"); RegionTimer reg(t);
+  // static Timer tstart("ruler3 - rule start");   
+  // static Timer tloop("ruler3 - rule loop"); 
 
+  // tstart.Start();
   float err, minerr, teterr, minteterr;
   char ok, found, hc;
   // vnetrule * rule;
@@ -75,23 +78,23 @@ int Meshing3 :: ApplyRules
   int loktestmode;
 
 
-  Array<int, PointIndex::BASE> pused;      // point is already mapped, number of uses
-  Array<char> fused;                       // face is already mapped
-  Array<PointIndex> pmap;                  // map of reference point to local point
-  Array<bool> pfixed;                      // point mapped by face-map
-  Array<int> fmapi;                        // face in reference is mapped to face nr ...
-  Array<int> fmapr;                        // face in reference is rotated to map 
-  Array<Point3d> transfreezone;            // transformed free-zone
+  NgArray<int, PointIndex::BASE> pused;      // point is already mapped, number of uses
+  NgArrayMem<char,100> fused;                       // face is already mapped
+  NgArrayMem<PointIndex,100> pmap;                  // map of reference point to local point
+  NgArrayMem<bool,100> pfixed;                      // point mapped by face-map
+  NgArrayMem<int,100> fmapi;                        // face in reference is mapped to face nr ...
+  NgArrayMem<int,100> fmapr;                        // face in reference is rotated to map 
+  NgArrayMem<Point3d,100> transfreezone;            // transformed free-zone
   INDEX_2_CLOSED_HASHTABLE<int> ledges(100); // edges in local environment
   
-  Array<Point3d> tempnewpoints;
-  Array<MiniElement2d> tempnewfaces;
-  Array<int> tempdelfaces;
-  Array<Element> tempelements;
-  Array<Box3d> triboxes;         // bounding boxes of local faces
+  NgArrayMem<Point3d,100> tempnewpoints;
+  NgArray<MiniElement2d> tempnewfaces;
+  NgArrayMem<int,100> tempdelfaces;
+  NgArray<Element> tempelements;
+  NgArrayMem<Box3d,100> triboxes;         // bounding boxes of local faces
 
-  Array<int, PointIndex::BASE> pnearness;
-  Array<int> fnearness;
+  NgArray<int, PointIndex::BASE> pnearness;
+  NgArray<int> fnearness;
 
   static int cnt = 0;
   cnt++;
@@ -221,7 +224,8 @@ int Meshing3 :: ApplyRules
 
 
   // check each rule:
-
+  // tstart.Stop();
+  // tloop.Start();
   for (int ri = 1; ri <= rules.Size(); ri++)
     {
       int base = (lfaces[0].GetNP() == 3) ? 100 : 200;
@@ -241,12 +245,12 @@ int Meshing3 :: ApplyRules
 	  if (rule->GetQuality() < 100) impossible = 0;
 
 	  if (testmode)
-	    sprintf (problems.Elem(ri), "Quality not ok");
+	    snprintf (problems.Elem(ri), 255, "Quality not ok");
 	  continue;
 	}
       
       if (testmode)
-	sprintf (problems.Elem(ri), "no mapping found");
+	snprintf (problems.Elem(ri), 255, "no mapping found");
       
       loktestmode = testmode || rule->TestFlag ('t') || tolerance > 5;
 
@@ -415,7 +419,7 @@ int Meshing3 :: ApplyRules
 	      if (loktestmode)
 		{
 		  (*testout) << "Faces Ok" << endl;
-		  sprintf (problems.Elem(ri), "Faces Ok");
+		  snprintf (problems.Elem(ri), 255, "Faces Ok");
 		}
 
 	      int npok = 1;
@@ -523,7 +527,7 @@ int Meshing3 :: ApplyRules
 			  for (auto pi : pmap)
 			    (*testout) << pi << " ";
 			  (*testout) << endl;
-			  sprintf (problems.Elem(ri), "mapping found");
+			  snprintf (problems.Elem(ri), 255, "mapping found");
 			  (*testout) << rule->GetNP(1) << " = " << lfaces.Get(1).GetNP() << endl;
 			}
 		      
@@ -638,7 +642,7 @@ int Meshing3 :: ApplyRules
 		      if (!rule->ConvexFreeZone())
 			{
 			  ok = 0;
-			  sprintf (problems.Elem(ri), "Freezone not convex");
+			  snprintf (problems.Elem(ri), 255, "Freezone not convex");
 
 			  if (loktestmode)
 			    (*testout) << "Freezone not convex" << endl;
@@ -646,7 +650,7 @@ int Meshing3 :: ApplyRules
 
 		      if (loktestmode)
 			{
-			  const Array<Point3d> & fz = rule->GetTransFreeZone();
+			  const NgArray<Point3d> & fz = rule->GetTransFreeZone();
 			  (*testout) << "Freezone: " << endl;
 			  for (int i = 1; i <= fz.Size(); i++)
 			    (*testout) << fz.Get(i) << endl;
@@ -669,7 +673,7 @@ int Meshing3 :: ApplyRules
 					{
 					  (*testout) << "Point " << i 
 						     << " in Freezone" << endl;
-					  sprintf (problems.Elem(ri), 
+					  snprintf (problems.Elem(ri), 255,
 						   "locpoint %d in Freezone", i);
 					}
 				      ok = 0;
@@ -681,7 +685,7 @@ int Meshing3 :: ApplyRules
 
 		      for (int i = 1; i <= lfaces.Size() && ok; i++)
 			{
-			  static Array<int> lpi(4);
+			  NgArrayMem<int, 10> lpi(4);
 
 			  if (!fused.Get(i))
 			    { 
@@ -783,7 +787,7 @@ int Meshing3 :: ApplyRules
 						     << lpoints[lfacei.PNum(4)] 
 						     << endl;
 
-				      sprintf (problems.Elem(ri), "triangle (%d, %d, %d) in Freezone",
+				      snprintf (problems.Elem(ri), 255, "triangle (%d, %d, %d) in Freezone",
 					       int(lfaces.Get(i).PNum(1)), 
 					       int(lfaces.Get(i).PNum(2)),
 					       int(lfaces.Get(i).PNum(3)));
@@ -826,7 +830,7 @@ int Meshing3 :: ApplyRules
 						     << lfaces.Get(i).PNum(2) << " - "
 						     << lfaces.Get(i).PNum(3) << endl;
 
-					  sprintf (problems.Elem(ri), "triangle (%d, %d, %d) in Freezone",
+					  snprintf (problems.Elem(ri), 255, "triangle (%d, %d, %d) in Freezone",
 						   int (lfaces.Get(i).PNum(1)), 
 						   int (lfaces.Get(i).PNum(2)),
 						   int (lfaces.Get(i).PNum(3)));
@@ -853,7 +857,7 @@ int Meshing3 :: ApplyRules
 			  if (loktestmode)
 			    {
 			      (*testout) << "Rule ok" << endl;
-			      sprintf (problems.Elem(ri), "Rule ok, err = %f", err);
+			      snprintf (problems.Elem(ri), 255, "Rule ok, err = %f", err);
 			    }
 
 
@@ -918,7 +922,7 @@ int Meshing3 :: ApplyRules
 				{
 				  if (loktestmode)
 				    {
-				      sprintf (problems.Elem(ri), "Orientation wrong");
+				      snprintf (problems.Elem(ri), 255, "Orientation wrong");
 				      (*testout) << "Orientation wrong ("<< n*v3 << ")" << endl;
 				    }
 				  ok = 0;
@@ -935,7 +939,7 @@ int Meshing3 :: ApplyRules
 				  {
 				    (*testout) << "Newpoint " << lpoints.Get(pmap.Get(i))
 					       << " outside convex hull" << endl;
-				    sprintf (problems.Elem(ri), "newpoint outside convex hull");
+				    snprintf (problems.Elem(ri), 255, "newpoint outside convex hull");
 				  }
 				ok = 0;
 				
@@ -1010,7 +1014,7 @@ int Meshing3 :: ApplyRules
 				{
 				  ok = 0;
 				  if (loktestmode)
-				    sprintf (problems.Elem(ri), "oldlen < newlen");
+				    snprintf (problems.Elem(ri), 255, "oldlen < newlen");
 				}
 			    }
 			  
@@ -1111,7 +1115,8 @@ int Meshing3 :: ApplyRules
       if (loktestmode)
 	(*testout) << "end rule" << endl;
     }
-
+  // tloop.Stop();
+  
   if (found)
     {
       /*
