@@ -5,10 +5,13 @@
 #include <csg.hpp>
 #endif
 #include <opti.hpp>
+#include <core/array.hpp>
+#include <core/taskmanager.hpp>
 
 
 namespace netgen
 {
+  using namespace ngcore;
   
 
   double MinFunctionSum :: Func (const Vector & x) const
@@ -88,7 +91,7 @@ namespace netgen
   }
 
   PointFunction1 :: PointFunction1 (Mesh::T_POINTS & apoints, 
-				    const Array<INDEX_3> & afaces,
+				    const NgArray<INDEX_3> & afaces,
 				    const MeshingParameters & amp,
 				    double ah)
     : points(apoints), faces(afaces), mp(amp)
@@ -176,12 +179,12 @@ namespace netgen
   class CheapPointFunction1 : public MinFunction
   {
     Mesh::T_POINTS & points;
-    const Array<INDEX_3> & faces;
+    const NgArray<INDEX_3> & faces;
     DenseMatrix m;
     double h;
   public:
     CheapPointFunction1 (Mesh::T_POINTS & apoints, 
-			 const Array<INDEX_3> & afaces,
+			 const NgArray<INDEX_3> & afaces,
 			 double ah);
   
     virtual double Func (const Vector & x) const;
@@ -189,7 +192,7 @@ namespace netgen
   };
 
   CheapPointFunction1 :: CheapPointFunction1 (Mesh::T_POINTS & apoints, 
-					      const Array<INDEX_3> & afaces,
+					      const NgArray<INDEX_3> & afaces,
 					      double ah)
     : points(apoints), faces(afaces)
   {
@@ -300,20 +303,21 @@ namespace netgen
   {
   public:
     Mesh::T_POINTS & points;
-    const Mesh::T_VOLELEMENTS & elements;
-    TABLE<int,PointIndex::BASE> elementsonpoint;
+    const Array<Element, ElementIndex> & elements;
+    Table<ElementIndex, PointIndex> &elementsonpoint;
+    bool own_elementsonpoint;
     const MeshingParameters & mp;
     PointIndex actpind;
     double h;
   
   public:
-    PointFunction (Mesh::T_POINTS & apoints, 
-		   const Mesh::T_VOLELEMENTS & aelements,
-		   const MeshingParameters & amp);
-    virtual ~PointFunction () { ; }
+    PointFunction (Mesh & mesh, const MeshingParameters & amp);
+    PointFunction (const PointFunction & pf);
+    virtual ~PointFunction () { if(own_elementsonpoint) delete &elementsonpoint; }
     virtual void SetPointIndex (PointIndex aactpind);
     void SetLocalH (double ah) { h = ah; }
     double GetLocalH () const { return h; }
+    const Table<ElementIndex, PointIndex> & GetPointToElementTable() { return elementsonpoint; };
     virtual double PointFunctionValue (const Point<3> & pp) const;
     virtual double PointFunctionValueGrad (const Point<3> & pp, Vec<3> & grad) const;
     virtual double PointFunctionValueDeriv (const Point<3> & pp, const Vec<3> & dir, double & deriv) const;
@@ -322,15 +326,41 @@ namespace netgen
   };
 
 
-  PointFunction :: PointFunction (Mesh::T_POINTS & apoints, 
-				  const Mesh::T_VOLELEMENTS & aelements,
-				  const MeshingParameters & amp)
-    : points(apoints), elements(aelements), elementsonpoint(apoints.Size()), mp(amp)
+  PointFunction :: PointFunction (const PointFunction & pf)
+    : points(pf.points), elements(pf.elements), elementsonpoint(pf.elementsonpoint), own_elementsonpoint(false), mp(pf.mp)
+  { }
+
+  PointFunction :: PointFunction (Mesh & mesh, const MeshingParameters & amp)
+    : points(mesh.Points()), elements(mesh.VolumeElements()), elementsonpoint(* new Table<ElementIndex,PointIndex>()), own_elementsonpoint(true), mp(amp)
   {
-    for (int i = 0; i < elements.Size(); i++)
-      if (elements[i].NP() == 4)
-        for (int j = 0; j < elements[i].NP(); j++)
-          elementsonpoint.Add (elements[i][j], i);  
+    static Timer tim("PointFunction - build elementsonpoint table"); RegionTimer reg(tim);
+
+    Array<bool, PointIndex> non_tet_points(points.Size());
+    non_tet_points = false;
+    // Don't optimize if point is adjacent to a non-tet element
+    ParallelForRange(elements.Range(), [&] (auto myrange)
+        {
+          for(auto ei : myrange)
+            {
+              const auto & el = elements[ei];
+              if(el.NP()!=4)
+                for(auto pi : el.PNums())
+                  non_tet_points[pi] = true;
+            }
+       });
+
+    elementsonpoint = ngcore::CreateSortedTable<ElementIndex, PointIndex>( elements.Range(),
+               [&](auto & table, ElementIndex ei)
+               {
+                 const auto & el = elements[ei];
+
+                 if(el.NP()!=4 || (mp.only3D_domain_nr && mp.only3D_domain_nr != el.GetIndex()) )
+                   return;
+
+                 for (PointIndex pi : el.PNums())
+                   if(!non_tet_points[pi])
+                     table.Add (pi, ei);
+               }, points.Size());
   }
 
   void PointFunction :: SetPointIndex (PointIndex aactpind)
@@ -348,9 +378,9 @@ namespace netgen
     hp = points[actpind];
     points[actpind] = Point<3> (pp);
 
-    for (int j = 0; j < elementsonpoint[actpind].Size(); j++)
+    for (auto ei : elementsonpoint[actpind])
       {
-        const Element & el = elements[elementsonpoint[actpind][j]];
+        const Element & el = elements[ei];
 	badness += CalcTetBadness (points[el[0]], points[el[1]], 
 				   points[el[2]], points[el[3]], -1, mp);
       }
@@ -368,9 +398,9 @@ namespace netgen
     Vec<3> vgradi, vgrad(0,0,0);
     points[actpind] = Point<3> (pp);
 
-    for (int j = 0; j < elementsonpoint[actpind].Size(); j++)
+    for (auto ei : elementsonpoint[actpind])
       {
-        const Element & el = elements[elementsonpoint[actpind][j]];
+        const Element & el = elements[ei];
 	for (int k = 0; k < 4; k++)
 	  if (el[k] == actpind)
 	    {
@@ -398,9 +428,9 @@ namespace netgen
     points[actpind] = pp;
     double f = 0;
 
-    for (int j = 0; j < elementsonpoint[actpind].Size(); j++)
+    for (auto ei : elementsonpoint[actpind])
       {
-        const Element & el = elements[elementsonpoint[actpind][j]];
+        const Element & el = elements[ei];
 
 	for (int k = 1; k <= 4; k++)
 	  if (el.PNum(k) == actpind)
@@ -422,12 +452,11 @@ namespace netgen
   int PointFunction :: MovePointToInner ()
   {
     // try point movement 
-    Array<Element2d> faces;
+    NgArray<Element2d> faces;
   
-    for (int j = 0; j < elementsonpoint[actpind].Size(); j++)
+    for (auto ei : elementsonpoint[actpind])
       {
-	const Element & el = 
-	  elements[elementsonpoint[actpind][j]];
+        const Element & el = elements[ei];
       
 	for (int k = 1; k <= 4; k++)
 	  if (el.PNum(k) == actpind)
@@ -476,19 +505,15 @@ namespace netgen
   {
     DenseMatrix m;
   public:
-    CheapPointFunction (Mesh::T_POINTS & apoints, 
-			const Mesh::T_VOLELEMENTS & aelements,
-			const MeshingParameters & amp);
+    CheapPointFunction (Mesh & mesh, const MeshingParameters & amp);
     virtual void SetPointIndex (PointIndex aactpind);
     virtual double PointFunctionValue (const Point<3> & pp) const;
     virtual double PointFunctionValueGrad (const Point<3> & pp, Vec<3> & grad) const;
   };
 
 
-  CheapPointFunction :: CheapPointFunction (Mesh::T_POINTS & apoints, 
-					    const Mesh::T_VOLELEMENTS & aelements,
-					    const MeshingParameters & amp)
-    : PointFunction (apoints, aelements, amp)
+  CheapPointFunction :: CheapPointFunction (Mesh & mesh, const MeshingParameters & amp)
+    : PointFunction (mesh, amp)
   {
     ;
   }
@@ -919,33 +944,6 @@ double Opti3EdgeMinFunction :: FuncGrad (const Vector & x, Vector & grad) const
 
 
 
-double CalcTotalBad (const Mesh::T_POINTS & points, 
-		     const Mesh::T_VOLELEMENTS & elements,
-		     const MeshingParameters & mp)
-{
-  double sum = 0;
-  double elbad;
-  
-  tets_in_qualclass.SetSize(20);
-  tets_in_qualclass = 0;
-
-  double teterrpow = mp.opterrpow;
-
-  for (int i = 1; i <= elements.Size(); i++)
-    {
-      elbad = pow (max2(CalcBad (points, elements.Get(i), 0, mp),1e-10),
-		   1/teterrpow);
-
-      int qualclass = int (20 / elbad + 1);
-      if (qualclass < 1) qualclass = 1;
-      if (qualclass > 20) qualclass = 20;
-      tets_in_qualclass.Elem(qualclass)++;
-
-      sum += elbad;
-    }
-  return sum;
-}
-
 int WrongOrientation (const Mesh::T_POINTS & points, const Element & el)
 {
   const Point3d & p1 = points[el.PNum(1)];
@@ -983,13 +981,13 @@ int WrongOrientation (const Mesh::T_POINTS & points, const Element & el)
 // {
 // public:
 //   Mesh::T_POINTS & points;
-//   const Mesh::T_VOLELEMENTS & elements;
+//   const NgArray<Element> & elements;
 //   TABLE<INDEX> elementsonpoint;
 //   PointIndex actpind;
   
 // public:
 //   JacobianPointFunction (Mesh::T_POINTS & apoints, 
-// 			 const Mesh::T_VOLELEMENTS & aelements);
+// 			 const NgArray<Element> & aelements);
   
 //   virtual void SetPointIndex (PointIndex aactpind);
 //   virtual double Func (const Vector & x) const;
@@ -1000,17 +998,12 @@ int WrongOrientation (const Mesh::T_POINTS & points, const Element & el)
 
 JacobianPointFunction :: 
 JacobianPointFunction (Mesh::T_POINTS & apoints, 
-		       const Mesh::T_VOLELEMENTS & aelements)
+		       const Array<Element, ElementIndex> & aelements)
   : points(apoints), elements(aelements), elementsonpoint(apoints.Size())
 {
-  INDEX i;
-  int j;
-  
-  for (i = 1; i <= elements.Size(); i++)
-    {
-      for (j = 1; j <= elements.Get(i).NP(); j++)
-	elementsonpoint.Add1 (elements.Get(i).PNum(j), i);  
-    }
+  for (int i = 0; i < elements.Size(); i++)
+      for (int j = 1; j <= elements[i].NP(); j++)
+          elementsonpoint.Add1 (elements[i].PNum(j), i+1);
 
   onplane = false;
 }
@@ -1023,24 +1016,20 @@ void JacobianPointFunction :: SetPointIndex (PointIndex aactpind)
 
 double JacobianPointFunction :: Func (const Vector & v) const
 {
-  int j;
   double badness = 0;
 
-  Point<3> hp = points.Elem(actpind);
+  Point<3> hp = points[actpind];
 
-  points.Elem(actpind) = hp + Vec<3> (v(0), v(1), v(2));
+  points[actpind] = hp + Vec<3> (v(0), v(1), v(2));
 
   if(onplane)
-    points.Elem(actpind) -= (v(0)*nv(0)+v(1)*nv(1)+v(2)*nv(2)) * nv;
+    points[actpind] -= (v(0)*nv(0)+v(1)*nv(1)+v(2)*nv(2)) * nv;
 
 
-  for (j = 1; j <= elementsonpoint.EntrySize(actpind); j++)
-    {
-      int eli = elementsonpoint.Get(actpind, j);
-      badness += elements.Get(eli).CalcJacobianBadness (points);
-    }
+  for (auto eli : elementsonpoint[actpind])
+      badness += elements[eli].CalcJacobianBadness (points);
   
-  points.Elem(actpind) = hp; 
+  points[actpind] = hp; 
 
   return badness;
 }
@@ -1052,25 +1041,24 @@ double JacobianPointFunction :: Func (const Vector & v) const
 double JacobianPointFunction :: 
 FuncGrad (const Vector & x, Vector & g) const
 {
-  int j, k;
+  int k;
   int lpi;
   double badness = 0;//, hbad;
 
-  Point<3> hp = points.Elem(actpind);
-  points.Elem(actpind) = hp + Vec<3> (x(0), x(1), x(2));
+  Point<3> hp = points[actpind];
+  points[actpind] = hp + Vec<3> (x(0), x(1), x(2));
 
   if(onplane)
-    points.Elem(actpind) -= (x(0)*nv(0)+x(1)*nv(1)+x(2)*nv(2)) * nv;
+    points[actpind] -= (x(0)*nv(0)+x(1)*nv(1)+x(2)*nv(2)) * nv;
 
   Vec<3> hderiv;
   //Vec3d vdir;
   g.SetSize(3);
   g = 0;
 
-  for (j = 1; j <= elementsonpoint.EntrySize(actpind); j++)
+  for (auto ei : elementsonpoint[actpind])
     {
-      int eli = elementsonpoint.Get(actpind, j);
-      const Element & el = elements.Get(eli);
+      const Element & el = elements[ei];
 
       lpi = 0;
       for (k = 1; k <= el.GetNP(); k++)
@@ -1078,7 +1066,7 @@ FuncGrad (const Vector & x, Vector & g) const
 	  lpi = k;
       if (!lpi) cerr << "loc point not found" << endl;
 
-      badness += elements.Get(eli).
+      badness += elements[ei].
 	CalcJacobianBadnessGradient (points, lpi, hderiv);
 
       for(k=0; k<3; k++)
@@ -1111,7 +1099,7 @@ FuncGrad (const Vector & x, Vector & g) const
   //(*testout) << "g = " << g << endl;
 
   
-  points.Elem(actpind) = hp; 
+  points[actpind] = hp; 
 
   return badness;
 }
@@ -1120,15 +1108,15 @@ FuncGrad (const Vector & x, Vector & g) const
 double JacobianPointFunction :: 
 FuncDeriv (const Vector & x, const Vector & dir, double & deriv) const
 {
-  int j, k;
+  int k;
   int lpi;
   double badness = 0;
 
-  Point<3> hp = points.Elem(actpind);
-  points.Elem(actpind) = Point<3> (hp + Vec3d (x(0), x(1), x(2)));
+  Point<3> hp = points[actpind];
+  points[actpind] = Point<3> (hp + Vec3d (x(0), x(1), x(2)));
 
   if(onplane)
-    points.Elem(actpind) -= (Vec3d (x(0), x(1), x(2))*nv) * nv;
+    points[actpind] -= (Vec3d (x(0), x(1), x(2))*nv) * nv;
 
   double hderiv;
   deriv = 0;
@@ -1140,10 +1128,9 @@ FuncDeriv (const Vector & x, const Vector & dir, double & deriv) const
       vdir -= scal*nv;
     }
 
-  for (j = 1; j <= elementsonpoint.EntrySize(actpind); j++)
+  for (auto ei : elementsonpoint[actpind])
     {
-      int eli = elementsonpoint.Get(actpind, j);
-      const Element & el = elements.Get(eli);
+      const Element & el = elements[ei];
 
       lpi = 0;
       for (k = 1; k <= el.GetNP(); k++)
@@ -1151,12 +1138,12 @@ FuncDeriv (const Vector & x, const Vector & dir, double & deriv) const
 	  lpi = k;
       if (!lpi) cerr << "loc point not found" << endl;
 
-      badness += elements.Get(eli).
+      badness += elements[ei].
 	CalcJacobianBadnessDirDeriv (points, lpi, vdir, hderiv);
       deriv += hderiv;
     }
   
-  points.Elem(actpind) = hp; 
+  points[actpind] = hp; 
 
   return badness;
   
@@ -1185,7 +1172,7 @@ void Mesh :: ImproveMesh (const CSG eometry & geometry, OPTIMIZEGOAL goal)
     }
 
   const char * savetask = multithread.task;
-  multithread.task = "Smooth Mesh";
+  multithread.task = "Optimize Volume: Smooth Mesh";
 
 
   TABLE<INDEX> surfelementsonpoint(points.Size());
@@ -1263,8 +1250,8 @@ void Mesh :: ImproveMesh (const CSG eometry & geometry, OPTIMIZEGOAL goal)
 
       //      if (uselocalh)
       {
-	double lh = GetH (points.Get(i));
-	pf->SetLocalH (GetH (points.Get(i)));
+	double lh = GetH(points.Get(i), points.Get(i).GetLayer());
+	pf->SetLocalH (lh);
 	par.typx = lh / 10;
 	//	  (*testout) << "lh(" << points.Get(i) << ") = " << lh << "\n";
       }
@@ -1353,77 +1340,63 @@ void Mesh :: ImproveMesh (const CSG eometry & geometry, OPTIMIZEGOAL goal)
   
 void Mesh :: ImproveMesh (const MeshingParameters & mp, OPTIMIZEGOAL goal)
 {
-  int typ = 1;
-  
+  static Timer t("Mesh::ImproveMesh"); RegionTimer reg(t);
+  static Timer tcoloring("coloring");
+  static Timer tcalcbadmax("Calc badmax");
+  static Timer topt("optimize");
+  static Timer trange("range");
+  static Timer tloch("loch");
+
+  BuildBoundaryEdges(false);
+
   (*testout) << "Improve Mesh" << "\n";
   PrintMessage (3, "ImproveMesh");
 
-  int np = GetNP();
+  // int np = GetNP();
   int ne = GetNE();
 
+  PointFunction pf_glob(*this, mp);
 
-  Array<double,PointIndex::BASE> perrs(np);
-  perrs = 1.0;
+  auto & elementsonpoint = pf_glob.GetPointToElementTable();
 
-  double bad1 = 0;
-  double badmax = 0;
+  const auto & getDofs = [&] (int i)
+  {
+      return elementsonpoint[i += PointIndex::BASE];
+  };
+
+  Array<int> colors(points.Size());
+
+  tcoloring.Start();
+  int ncolors = ngcore::ComputeColoring( colors, ne, getDofs );
+  auto color_table = CreateTable<PointIndex, int>( points.Size(),
+         [&] ( auto & table, int i )
+          {
+            PointIndex pi = i+static_cast<int>(PointIndex::BASE);
+            table.Add(colors[i], pi);
+          }, ncolors);
+
+  tcoloring.Stop();
 
   if (goal == OPT_QUALITY)
     {
-      for (int i = 1; i <= ne; i++)
-	{
-	  const Element & el = VolumeElement(i);
-	  if (el.GetType() != TET)
-	    continue;
-	  
-	  double hbad = CalcBad (points, el, 0, mp);
-	  for (int j = 0; j < 4; j++)
-	    perrs[el[j]] += hbad;
-	  
-	  bad1 += hbad;
-	}
-      
-      for (int i = perrs.Begin(); i < perrs.End(); i++)
-	if (perrs[i] > badmax) 
-	  badmax = perrs[i];
-      badmax = 0;
-    }
-
-  if (goal == OPT_QUALITY)
-    {
-      bad1 = CalcTotalBad (points, volelements, mp);
+      double bad1 = CalcTotalBad (mp);
       (*testout) << "Total badness = " << bad1 << endl;
       PrintMessage (5, "Total badness = ", bad1);
     }
-  
-  Vector x(3);
-  
+
+
   (*testout) << setprecision(8);
-  
-  //int uselocalh = mparam.uselocalh;
 
+  Array<double, PointIndex> pointh (points.Size());
 
-  PointFunction * pf;
-
-  if (typ == 1)
-    pf = new PointFunction(points, volelements, mp);
-  else
-    pf = new CheapPointFunction(points, volelements, mp);
-
-  //  pf->SetLocalH (h);
-  
-  Opti3FreeMinFunction freeminf(*pf);
-
-  OptiParameters par;
-  par.maxit_linsearch = 20;
-  par.maxit_bfgs = 20;
-
-  Array<double, PointIndex::BASE> pointh (points.Size());
-
-  if(lochfunc)
+  if(HasLocalHFunction())
     {
-      for (PointIndex pi : points.Range())
-	pointh[pi] = GetH(points[pi]);
+      RegionTimer rt(tloch);
+      ParallelForRange(points.Range(), [&] (auto myrange)
+         {
+           for(auto pi : myrange)
+             pointh[pi] = GetH(pi);
+         });
     }
   else
     {
@@ -1436,73 +1409,73 @@ void Mesh :: ImproveMesh (const MeshingParameters & mp, OPTIMIZEGOAL goal)
               pointh[pi] = h;
 	}
     }
- 
-
-  int printmod = 1;
-  char printdot = '.';
-  if (points.Size() > 1000)
-    {
-      printmod = 10;
-      printdot = '+';
-    }
-  if (points.Size() > 10000)
-    {
-      printmod = 100;
-      printdot = '*';
-    }
-
 
   const char * savetask = multithread.task;
-  multithread.task = "Smooth Mesh";
+  multithread.task = "Optimize Volume: Smooth Mesh";
 
-  for (PointIndex pi : points.Range())
-    if ( (*this)[pi].Type() == INNERPOINT && perrs[pi] > 0.01 * badmax)
+  topt.Start();
+  for (auto icolor : Range(ncolors))
+  {
+      if (multithread.terminate)
+          throw NgException ("Meshing stopped");
+
+      ParallelForRange( color_table[icolor].Range(), [&](auto myrange)
       {
-	if (multithread.terminate)
-	  throw NgException ("Meshing stopped");
+        RegionTracer reg(ngcore::TaskManager::GetThreadId(), trange, myrange.Size());
+        Vector x(3);
 
-	multithread.percent = 100.0 * (pi+1-PointIndex::BASE) / points.Size();
+        PointFunction pf{pf_glob};
 
-        if (  (pi+1-PointIndex::BASE) % printmod == 0) PrintDot (printdot);
+        Opti3FreeMinFunction freeminf(pf);
 
-	double lh = pointh[pi];
-	pf->SetLocalH (lh);
-	par.typx = lh;
+        OptiParameters par;
+        par.maxit_linsearch = 20;
+        par.maxit_bfgs = 20;
 
-	freeminf.SetPoint (points[pi]);
-	pf->SetPointIndex (pi);
+        for (auto i : myrange)
+        {
+          PointIndex pi = color_table[icolor][i];
+          if ( (*this)[pi].Type() == INNERPOINT )
+          {
+            double lh = pointh[pi];
+            pf.SetLocalH (lh);
+            par.typx = lh;
 
-	x = 0;
-	int pok;
-	pok = freeminf.Func (x) < 1e10; 
+            freeminf.SetPoint (points[pi]);
+            pf.SetPointIndex (pi);
 
-	if (!pok)
-	  {
-	    pok = pf->MovePointToInner ();
+            x = 0;
+            int pok;
+            pok = freeminf.Func (x) < 1e10;
 
-	    freeminf.SetPoint (points[pi]);
-	    pf->SetPointIndex (pi);
-	  }
+            if (!pok)
+              {
+                pok = pf.MovePointToInner ();
 
-	if (pok)
-	  {
-            //*testout << "start BFGS, pok" << endl;
-	    BFGS (x, freeminf, par);
-            //*testout << "BFGS complete, pok" << endl;
-	    points[pi](0) += x(0);
-	    points[pi](1) += x(1);
-	    points[pi](2) += x(2);
-	  }
-      }
-  PrintDot ('\n');
-  
-  delete pf;
+                freeminf.SetPoint (points[pi]);
+                pf.SetPointIndex (pi);
+              }
+
+            if (pok)
+              {
+                //*testout << "start BFGS, pok" << endl;
+                BFGS (x, freeminf, par);
+                //*testout << "BFGS complete, pok" << endl;
+                points[pi](0) += x(0);
+                points[pi](1) += x(1);
+                points[pi](2) += x(2);
+              }
+          }
+        }
+      }, 4*ngcore::TaskManager::GetNumThreads());
+  }
+  topt.Stop();
 
   multithread.task = savetask;
 
   if (goal == OPT_QUALITY)
     {
-      bad1 = CalcTotalBad (points, volelements, mp);
+      double bad1 = CalcTotalBad (mp);
       (*testout) << "Total badness = " << bad1 << endl;
       PrintMessage (5, "Total badness = ", bad1);
     }
@@ -1510,12 +1483,11 @@ void Mesh :: ImproveMesh (const MeshingParameters & mp, OPTIMIZEGOAL goal)
 
 
 
-
 // Improve Condition number of Jacobian, any elements  
 void Mesh :: ImproveMeshJacobian (const MeshingParameters & mp,
-				  OPTIMIZEGOAL goal, const BitArray * usepoint)
+				  OPTIMIZEGOAL goal, const NgBitArray * usepoint)
 {
-  int i, j;
+  // int i, j;
   
   (*testout) << "Improve Mesh Jacobian" << "\n";
   PrintMessage (3, "ImproveMesh Jacobian");
@@ -1535,33 +1507,34 @@ void Mesh :: ImproveMeshJacobian (const MeshingParameters & mp,
   par.maxit_linsearch = 20;
   par.maxit_bfgs = 20;
   
-  BitArray badnodes(np);
+  NgBitArray badnodes(np);
   badnodes.Clear();
 
-  for (i = 1; i <= ne; i++)
+  for (int i = 1; i <= ne; i++)
     {
       const Element & el = VolumeElement(i);
       double bad = el.CalcJacobianBadness (Points());
       if (bad > 1)
-	for (j = 1; j <= el.GetNP(); j++)
+	for (int j = 1; j <= el.GetNP(); j++)
 	  badnodes.Set (el.PNum(j));
     }
 
-  Array<double, PointIndex::BASE> pointh (points.Size());
+  NgArray<double, PointIndex::BASE, PointIndex> pointh (points.Size());
 
-  if(lochfunc)
+  if(HasLocalHFunction())
     {
-      for(i = 1; i<=points.Size(); i++)
-	pointh[i] = GetH(points.Get(i));
+      // for(i = 1; i<=points.Size(); i++)
+      for (PointIndex pi : points.Range())
+	pointh[pi] = GetH(pi);
     }
   else
     {
       pointh = 0;
-      for(i=0; i<GetNE(); i++)
+      for (int i=0; i<GetNE(); i++)
 	{
 	  const Element & el = VolumeElement(i+1);
 	  double h = pow(el.Volume(points),1./3.);
-	  for(j=1; j<=el.GetNV(); j++)
+	  for(int j=1; j<=el.GetNV(); j++)
 	    if(h > pointh[el.PNum(j)])
 	      pointh[el.PNum(j)] = h;
 	}
@@ -1570,19 +1543,20 @@ void Mesh :: ImproveMeshJacobian (const MeshingParameters & mp,
 
 
   const char * savetask = multithread.task;
-  multithread.task = "Smooth Mesh Jacobian";
+  multithread.task = "Optimize Volume: Smooth Mesh Jacobian";
   
-  for (PointIndex pi = points.Begin(); i < points.End(); pi++)
+  // for (PointIndex pi = points.Begin(); i < points.End(); pi++)
+  for (PointIndex pi : points.Range())
     {
       if ((*this)[pi].Type() != INNERPOINT)
 	continue;
 
-      if(usepoint && !usepoint->Test(i))
+      if(usepoint && !usepoint->Test(pi))
 	continue;
 
       //(*testout) << "improvejac, p = " << i << endl;
 
-      if (goal == OPT_WORSTCASE && !badnodes.Test(i))
+      if (goal == OPT_WORSTCASE && !badnodes.Test(pi))
 	continue;
       //	(*testout) << "smooth p " << i << endl;
 
@@ -1593,15 +1567,15 @@ void Mesh :: ImproveMeshJacobian (const MeshingParameters & mp,
       if (multithread.terminate)
 	throw NgException ("Meshing stopped");
 
-      multithread.percent = 100.0 * i / points.Size();
+      multithread.percent = 100.0 * pi / points.Size();
 
       if (points.Size() < 1000)
 	PrintDot ();
       else
-	if (i % 10 == 0)
+	if (pi % 10 == 0)
 	  PrintDot ('+');
 
-      double lh = pointh[i];
+      double lh = pointh[pi];
       par.typx = lh;
 
       pf.SetPointIndex (pi);
@@ -1614,9 +1588,9 @@ void Mesh :: ImproveMeshJacobian (const MeshingParameters & mp,
           //*testout << "start BFGS, Jacobian" << endl;
 	  BFGS (x, pf, par);
           //*testout << "end BFGS, Jacobian" << endl;
-	  points.Elem(i)(0) += x(0);
-	  points.Elem(i)(1) += x(1);
-	  points.Elem(i)(2) += x(2);
+	  points[pi](0) += x(0);
+	  points[pi](1) += x(1);
+	  points[pi](2) += x(2);
 	}
       else
 	{
@@ -1634,12 +1608,12 @@ void Mesh :: ImproveMeshJacobian (const MeshingParameters & mp,
 
 // Improve Condition number of Jacobian, any elements  
 void Mesh :: ImproveMeshJacobianOnSurface (const MeshingParameters & mp,
-					   const BitArray & usepoint, 
-					   const Array< Vec<3>* > & nv,
+					   const NgBitArray & usepoint, 
+					   const NgArray< Vec<3>* > & nv,
 					   OPTIMIZEGOAL goal,
-					   const Array< Array<int,PointIndex::BASE>* > * idmaps)
+					   const NgArray< NgArray<int,PointIndex::BASE>* > * idmaps)
 {
-  int i, j;
+  // int i, j;
   
   (*testout) << "Improve Mesh Jacobian" << "\n";
   PrintMessage (3, "ImproveMesh Jacobian");
@@ -1654,8 +1628,8 @@ void Mesh :: ImproveMeshJacobianOnSurface (const MeshingParameters & mp,
   
   JacobianPointFunction pf(points, volelements);
 
-  Array< Array<int,PointIndex::BASE>* > locidmaps;
-  const Array< Array<int,PointIndex::BASE>* > * used_idmaps;
+  NgArray< NgArray<int,PointIndex::BASE>* > locidmaps;
+  const NgArray< NgArray<int,PointIndex::BASE>* > * used_idmaps;
 
   if(idmaps)
     used_idmaps = idmaps;
@@ -1663,11 +1637,11 @@ void Mesh :: ImproveMeshJacobianOnSurface (const MeshingParameters & mp,
     {
       used_idmaps = &locidmaps;
       
-      for(i=1; i<=GetIdentifications().GetMaxNr(); i++)
+      for(int i=1; i<=GetIdentifications().GetMaxNr(); i++)
 	{
 	  if(GetIdentifications().GetType(i) == Identifications::PERIODIC)
 	    {
-	      locidmaps.Append(new Array<int,PointIndex::BASE>);
+	      locidmaps.Append(new NgArray<int,PointIndex::BASE>);
 	      GetIdentifications().GetMap(i,*locidmaps.Last(),true);
 	    }
 	}
@@ -1690,33 +1664,34 @@ void Mesh :: ImproveMeshJacobianOnSurface (const MeshingParameters & mp,
   par.maxit_linsearch = 20;
   par.maxit_bfgs = 20;
   
-  BitArray badnodes(np);
+  NgBitArray badnodes(np);
   badnodes.Clear();
 
-  for (i = 1; i <= ne; i++)
+  for (int i = 1; i <= ne; i++)
     {
       const Element & el = VolumeElement(i);
       double bad = el.CalcJacobianBadness (Points());
       if (bad > 1)
-	for (j = 1; j <= el.GetNP(); j++)
+	for (int j = 1; j <= el.GetNP(); j++)
 	  badnodes.Set (el.PNum(j));
     }
 
-  Array<double, PointIndex::BASE> pointh (points.Size());
+  NgArray<double, PointIndex::BASE> pointh (points.Size());
  
-  if(lochfunc)
+  if(HasLocalHFunction())
     {
-      for(i=1; i<=points.Size(); i++)
-	pointh[i] = GetH(points.Get(i));
+      // for(i=1; i<=points.Size(); i++)
+      for (PointIndex pi : points.Range())
+	pointh[pi] = GetH(pi);
     }
   else
     {
       pointh = 0;
-      for(i=0; i<GetNE(); i++)
+      for(int i=0; i<GetNE(); i++)
 	{
 	  const Element & el = VolumeElement(i+1);
 	  double h = pow(el.Volume(points),1./3.);
-	  for(j=1; j<=el.GetNV(); j++)
+	  for(int j=1; j<=el.GetNV(); j++)
 	    if(h > pointh[el.PNum(j)])
 	      pointh[el.PNum(j)] = h;
 	}
@@ -1724,14 +1699,15 @@ void Mesh :: ImproveMeshJacobianOnSurface (const MeshingParameters & mp,
 
 
   const char * savetask = multithread.task;
-  multithread.task = "Smooth Mesh Jacobian";
+  multithread.task = "Optimize Volume: Smooth Mesh Jacobian";
   
-  for (PointIndex pi = points.Begin(); pi <= points.End(); pi++)
-    if ( usepoint.Test(i) )
+  // for (PointIndex pi = points.Begin(); pi <= points.End(); pi++)
+  for (PointIndex pi : points.Range())
+    if ( usepoint.Test(pi) )
       {
 	//(*testout) << "improvejac, p = " << i << endl;
 
-	if (goal == OPT_WORSTCASE && !badnodes.Test(i))
+	if (goal == OPT_WORSTCASE && !badnodes.Test(pi))
 	  continue;
 	//	(*testout) << "smooth p " << i << endl;
 
@@ -1742,15 +1718,15 @@ void Mesh :: ImproveMeshJacobianOnSurface (const MeshingParameters & mp,
 	if (multithread.terminate)
 	  throw NgException ("Meshing stopped");
 
-	multithread.percent = 100.0 * i / points.Size();
+	multithread.percent = 100.0 * pi / points.Size();
 
 	if (points.Size() < 1000)
 	  PrintDot ();
 	else
-	  if (i % 10 == 0)
+	  if (pi % 10 == 0)
 	    PrintDot ('+');
 
-	double lh = pointh[i];//GetH(points.Get(i));
+	double lh = pointh[pi];//GetH(points.Get(i));
 	par.typx = lh;
 
 	pf.SetPointIndex (pi);
@@ -1758,29 +1734,29 @@ void Mesh :: ImproveMeshJacobianOnSurface (const MeshingParameters & mp,
 	PointIndex brother (-1);
 	if(usesum)
 	  {
-	    for(j=0; brother == -1 && j<used_idmaps->Size(); j++)
+	    for(int j=0; brother == -1 && j<used_idmaps->Size(); j++)
 	      {
-		if(i < (*used_idmaps)[j]->Size() + PointIndex::BASE)
+		if(pi < (*used_idmaps)[j]->Size() + PointIndex::BASE)
 		  {
-		    brother = (*(*used_idmaps)[j])[i];
-		    if(brother == i || brother == 0)
+		    brother = (*(*used_idmaps)[j])[pi];
+		    if(brother == pi || brother == 0)
 		      brother = -1;
 		  }
 	      }
-	    if(brother >= i)
+	    if(brother >= pi)
 	      {
 		pf2ptr->SetPointIndex(brother);
 		pf2ptr->SetNV(*nv[brother-1]);
 	      }
 	  }
 
-	if(usesum && brother < i)
+	if(usesum && brother < pi)
 	  continue;
 
 	//pf.UnSetNV(); x = 0;
 	//(*testout) << "before " << pf.Func(x);
 
-	pf.SetNV(*nv[i-1]);
+	pf.SetNV(*nv[pi-1]);
 
 	x = 0;
 	int pok = (brother == -1) ? (pf.Func (x) < 1e10) : (pf_sum.Func (x) < 1e10);
@@ -1794,12 +1770,12 @@ void Mesh :: ImproveMeshJacobianOnSurface (const MeshingParameters & mp,
 	      BFGS (x, pf_sum, par);
 
 
-	    for(j=0; j<3; j++)
-	      points.Elem(i)(j) += x(j);// - scal*nv[i-1].X(j);
+	    for(int j=0; j<3; j++)
+	      points[pi](j) += x(j);// - scal*nv[i-1].X(j);
 
 	    if(brother != -1)
-	      for(j=0; j<3; j++)
-		points.Elem(brother)(j) += x(j);// - scal*nv[brother-1].X(j);
+	      for(int j=0; j<3; j++)
+		points[brother](j) += x(j);// - scal*nv[brother-1].X(j);
 
 
 	  }
@@ -1817,7 +1793,7 @@ void Mesh :: ImproveMeshJacobianOnSurface (const MeshingParameters & mp,
   PrintDot ('\n');
 
   delete pf2ptr;
-  for(i=0; i<locidmaps.Size(); i++)
+  for(int i=0; i<locidmaps.Size(); i++)
     delete locidmaps[i];
 
   multithread.task = savetask;
