@@ -14,6 +14,7 @@
 #include "occgeom.hpp"
 #include "Partition_Spliter.hxx"
 
+#include <BinTools.hxx>
 #include <BOPAlgo_Builder.hxx>
 #include <BRepBndLib.hxx>
 #include <BRepBuilderAPI_Copy.hxx>
@@ -23,7 +24,6 @@
 #include <BRepExtrema_DistShapeShape.hxx>
 #include <BRepGProp.hxx>
 #include <BRepLib.hxx>
-#include <BRepMesh_IncrementalMesh.hxx>
 #include <BRepOffsetAPI_Sewing.hxx>
 #include <BRepTools.hxx>
 #include <IGESCAFControl_Reader.hxx>
@@ -200,6 +200,11 @@ namespace netgen
       return *faces[fmap.FindIndex(shape)-1];
   }
 
+  const GeometrySolid & OCCGeometry :: GetSolid(const TopoDS_Shape & shape) const
+  {
+      return *solids[somap.FindIndex(shape)-1];
+  }
+
 
   string STEP_GetEntityName(const TopoDS_Shape & theShape, STEPCAFControl_Reader * aReader)
   {
@@ -416,16 +421,7 @@ namespace netgen
       }
 #endif
 
-#ifdef OCC_HAVE_HISTORY    
-    Handle(BRepTools_History) history = aBuilder.History ();
-    
-    for (TopExp_Explorer e(shape, TopAbs_SOLID); e.More(); e.Next())
-      {
-        if (auto name = OCCGeometry::GetProperties(e.Current()).name)
-          for (auto mods : history->Modified(e.Current()))
-            OCCGeometry::GetProperties(mods).name = *name;
-      }
-#endif // OCC_HAVE_HISTORY    
+    PropagateProperties(aBuilder, shape);
     
     // result of the operation
     shape = aBuilder.Shape();
@@ -1136,6 +1132,21 @@ namespace netgen
          }
       }
       */
+
+      std::map<int, ArrayMem<int, 10>> free_edges_in_solid;
+      for(auto i1 : Range(1, somap.Extent()+1))
+      {
+          auto s = somap(i1);
+          for (auto edge : MyExplorer(s, TopAbs_EDGE, TopAbs_WIRE))
+            if (!emap.Contains(edge))
+              {
+                free_edges_in_solid[i1].Append(emap.Add (edge));
+                for (auto vertex : MyExplorer(edge, TopAbs_VERTEX))
+                  if (!vmap.Contains(vertex))
+                    vmap.Add (vertex);
+              }
+      }
+
       for (auto edge : MyExplorer(shape, TopAbs_EDGE, TopAbs_WIRE))
         if (!emap.Contains(edge))
           {
@@ -1248,11 +1259,11 @@ namespace netgen
           auto occ_solid = make_unique<OCCSolid>(s);
           if(HaveProperties(s))
             occ_solid->properties = GetProperties(s);
-          solids.Append(std::move(occ_solid));
-
           for(auto f : GetFaces(s))
           {
               auto & face = static_cast<OCCFace&>(GetFace(f));
+              face.properties.maxh = min2(face.properties.maxh, occ_solid->properties.maxh);
+
               if(face.domin==-1)
                   face.domin = k;
               else
@@ -1260,8 +1271,25 @@ namespace netgen
               if(face.Shape().Orientation() == TopAbs_INTERNAL)
                   face.domout = k;
           }
+
+          if(free_edges_in_solid.count(i1))
+            for(auto ei : free_edges_in_solid[i1])
+            {
+              auto & edge = GetEdge(emap(ei));
+              edge.properties.maxh = min(edge.properties.maxh, occ_solid->properties.maxh);
+              edge.domin = k;
+              edge.domout = k;
+              occ_solid->free_edges.Append(&GetEdge(emap(ei)));
+            }
+          solids.Append(std::move(occ_solid));
       }
 
+      // Propagate maxh to children
+      for(auto& face : faces)
+        for(auto& edge : face->edges)
+          edge->properties.maxh = min2(edge->properties.maxh,
+                                       face->properties.maxh);
+
       // Add identifications
       auto add_identifications = [&](auto & shapes, auto & shape_map)
       {
@@ -1387,12 +1415,9 @@ namespace netgen
 
    void OCCGeometry :: BuildVisualizationMesh (double deflection)
    {
-      cout << "Preparing visualization (deflection = " << deflection << ") ... " << flush;
-
-      BRepTools::Clean (shape);
-      // BRepMesh_IncrementalMesh::
-      BRepMesh_IncrementalMesh (shape, deflection, true);
-      cout << "done" << endl;
+      // cout << IM(5) << "Preparing visualization (deflection = " << deflection << ") ... " << flush;
+      BuildTriangulation(shape);
+      // cout << IM(5) << "done" << endl;
    }
 
 
@@ -1624,8 +1649,12 @@ namespace netgen
 
       if(!result)
       {
-         delete occgeo;
-         return NULL;
+        result = BinTools::Read(occgeo->shape, filename.string().c_str());
+        if (!result)
+          {
+            delete occgeo;
+            throw Exception("Could not read BREP file " + filename.string());
+          }
       }
 
       occgeo->changed = 1;
@@ -1714,11 +1743,6 @@ namespace netgen
         BRepTools::Read(shape, ss, builder);
       }
 
-    // enumerate shapes and archive only integers
-    auto my_hash = [](const TopoDS_Shape & key) {
-        auto occ_hash = key.HashCode(1<<31UL);
-        return std::hash<decltype(occ_hash)>()(occ_hash);
-    };
     TopTools_IndexedMapOfShape shape_map;
     Array<TopoDS_Shape> shape_list;
 
@@ -1758,7 +1782,18 @@ namespace netgen
                   id_from = shape_map.FindIndex(id.from)-1;
                   id_to = shape_map.FindIndex(id.to)-1;
                 }
-                ar & id_from & id_to & id.trafo & id.name;
+                ar & id_from & id_to;
+
+                // trafo is now optional -> special treatment necessary for backward compatibility
+                if(ar.Output() || netgen_version >= "v6.2.2403-34-g571cbbe4")
+                  ar & id.trafo;
+                else
+                {
+                  auto trafo = Transformation<3>();
+                  ar & trafo;
+                  id.trafo = trafo;
+                }
+                ar & id.name;
                 if(ar.Input())
                 {
                     id.from = shape_list[id_from];
@@ -1786,7 +1821,7 @@ namespace netgen
    "Face", "Wire", "Edge", "Vertex"};
 
   const char * orientationstring[] =
-     {"+", "-"};
+    {"+", "-", "i", "e"};
 
 
 
@@ -1845,6 +1880,16 @@ namespace netgen
          }
 
          str << "{" << shapename[l] << " " << count2;
+         if(HaveProperties(e.Current()))
+           {
+             const auto& props = GetProperties(e.Current());
+             if(props.name || props.maxh < 1e99)
+               str << " - ";
+             if(props.name)
+               str << props.GetName();
+             if(props.maxh < 1e99)
+               str << " maxh(" << props.maxh << ")";
+           }
 
          if (l <= TopAbs_EDGE)
          {
@@ -2340,21 +2385,27 @@ namespace netgen
 
           for(auto & ident : identifications)
           {
+              const auto& to = STEPConstruct::FindEntity(finder, ident.from == shape ? ident.to : ident.from);
+              if(to.IsNull())
+                  continue;
               Array<Handle(StepRepr_RepresentationItem)> items;
-              // items.Append(STEPConstruct::FindEntity(finder, ident.other)); // TODO!
-              auto & m = ident.trafo.GetMatrix();
+              items.Append(MakeReal(ident.from == shape ? 1 : 0));
+              items.Append(to);
+              Transformation<3> trafo;
+              if(ident.trafo) trafo = *ident.trafo;
+              auto & m = trafo.GetMatrix();
               for(auto i : Range(9))
                   items.Append(MakeReal(m(i)));
-              auto & v = ident.trafo.GetVector();
+              auto & v = trafo.GetVector();
               for(auto i : Range(3))
                   items.Append(MakeReal(v(i)));
-              for(auto & item : items.Range(1,items.Size()))
+              items.Append(MakeInt(ident.type));
+              for(auto & item : items.Range(0, items.Size()))
                   model->AddEntity(item);
               ident_items.Append(MakeCompound(items, ident.name));
           }
-
-          for(auto & item : ident_items.Range(1,ident_items.Size()))
-              model->AddEntity(item);
+          for(auto & item : ident_items.Range(1, ident_items.Size()))
+            model->AddEntity(item);
           auto comp = MakeCompound(ident_items, "netgen_geometry_identification");
           model->AddEntity(comp);
       }
@@ -2371,15 +2422,29 @@ namespace netgen
               auto id_item = Handle(StepRepr_CompoundRepresentationItem)::DownCast(idents->ItemElementValue(i));
               OCCIdentification ident;
               ident.name = id_item->Name()->ToCString();
-              // ident.other = TransferBRep::ShapeResult(transProc->Find(id_item->ItemElementValue(1))); /TODO!
+              auto is_from = ReadReal(id_item->ItemElementValue(1));
+              if(is_from)
+                {
+                  ident.from = shape_origin;
+                  ident.to = TransferBRep::ShapeResult(transProc->Find(id_item->ItemElementValue(2)));
+                }
+              else
+                {
+                  ident.from = TransferBRep::ShapeResult(
+                      transProc->Find(id_item->ItemElementValue(2)));
+                  ident.to = shape_origin;
+                }
 
-              auto & m = ident.trafo.GetMatrix();
+              Transformation<3> trafo;
+              auto & m = trafo.GetMatrix();
               for(auto i : Range(9))
                   m(i) = ReadReal(id_item->ItemElementValue(3+i));
-              auto & v = ident.trafo.GetVector();
+              auto & v = trafo.GetVector();
               for(auto i : Range(3))
                   v(i) = ReadReal(id_item->ItemElementValue(12+i));
-
+              if(FlatVector(9, &trafo.GetMatrix()(0,0)).L2Norm() != .0 && trafo.GetVector().Length2() != .0)
+                  ident.trafo = trafo;
+              ident.type = Identifications::ID_TYPE(ReadInt(id_item->ItemElementValue(15)));
               result.push_back(ident);
           }
           OCCGeometry::GetIdentifications(shape_origin) = result;
