@@ -9,6 +9,7 @@
 #include <meshing.hpp>
 
 #include "occgeom.hpp"
+#include "occ_utils.hpp"
 
 #pragma clang diagnostic push
 #pragma clang diagnostic ignored "-Wdeprecated-declarations"
@@ -21,6 +22,7 @@
 #include <BRepAlgo_NormalProjection.hxx>
 #include <BRepBuilderAPI_MakeEdge.hxx>
 #include <BRepBuilderAPI_MakeFace.hxx>
+#include <BRepBuilderAPI_MakePolygon.hxx>
 #include <BRepBuilderAPI_MakeVertex.hxx>
 #include <BRepBuilderAPI_MakeWire.hxx>
 #include <BRepBuilderAPI_Transform.hxx>
@@ -32,7 +34,6 @@
 #include <BRepGProp.hxx>
 #include <BRepLProp_SLProps.hxx>
 #include <BRepLib.hxx>
-#include <BRepMesh_IncrementalMesh.hxx>
 #include <BRepOffsetAPI_MakeOffset.hxx>
 #include <BRepOffsetAPI_MakePipe.hxx>
 #include <BRepOffsetAPI_MakePipeShell.hxx>
@@ -45,7 +46,6 @@
 #include <BRepPrimAPI_MakePrism.hxx>
 #include <BRepPrimAPI_MakeRevol.hxx>
 #include <BRepPrimAPI_MakeSphere.hxx>
-#include <BRepTools.hxx>
 #include <GCE2d_MakeArcOfCircle.hxx>
 #include <GCE2d_MakeCircle.hxx>
 #include <GCE2d_MakeSegment.hxx>
@@ -610,11 +610,11 @@ public:
     return Circle (pos.X(), pos.Y(), r);
   }
   
-  shared_ptr<WorkPlane> Close ()
+  shared_ptr<WorkPlane> Close (optional<string> name = nullopt)
   {
     if (startpnt.Distance(localpos.Location()) > 1e-10)
       {
-        LineTo (startpnt.X(), startpnt.Y());
+        LineTo (startpnt.X(), startpnt.Y(), name);
         return shared_from_this();                    
       }
 
@@ -631,13 +631,18 @@ public:
   
   auto Offset(double d)
   {
+    Finish();
     TopoDS_Wire wire = wires.back();
     wires.pop_back();
-    BRepOffsetAPI_MakeOffset builder;
-    builder.AddWire(wire);
+
+    // handle wires containing a single edge correctly, see
+    // https://dev.opencascade.org/content/brepoffsetapimakeoffset-open-topodswire
+    BRepBuilderAPI_MakeFace makeFace{gp_Pln{axes}};
+    makeFace.Add(wire);
+    BRepOffsetAPI_MakeOffset builder(makeFace.Face());
     builder.Perform(d);
     auto shape = builder.Shape();
-    wires.push_back (TopoDS::Wire(shape.Reversed()));
+    wires.push_back (TopoDS::Wire(shape));
     return shared_from_this();
   }
   
@@ -679,6 +684,10 @@ DLL_HEADER void ExportNgOCCShapes(py::module &m)
     .export_values()
     ;
   
+  m.def("ResetGlobalShapeProperties", [] () {
+    OCCGeometry::global_shape_properties.clear();
+    OCCGeometry::global_shape_property_indices.Clear();
+  });
   
   py::class_<TopoDS_Shape> (m, "TopoDS_Shape")
     .def("__str__", [] (const TopoDS_Shape & shape)
@@ -736,6 +745,10 @@ DLL_HEADER void ExportNgOCCShapes(py::module &m)
            return Mass(shape);
       }, "returns mass of shape, what is length, face, or volume")
 
+    .def_property_readonly("inertia", [](const TopoDS_Shape & shape) {
+           return Properties(shape).MatrixOfInertia();           
+      }, "returns matrix of inertia of shape")
+    
     .def("Move", [](const TopoDS_Shape & shape, const gp_Vec v)
          {
            // which one to choose ? 
@@ -768,7 +781,7 @@ DLL_HEADER void ExportNgOCCShapes(py::module &m)
            PropagateProperties(builder, shape, occ2ng(trafo));
            return builder.Shape();
          }, py::arg("axes"),
-         "copy shape, and mirror over plane defined by 'axes'")
+         "copy shape, and mirror over XY - plane defined by 'axes'")
     
     .def("Mirror", [] (const TopoDS_Shape & shape, const gp_Ax1 & ax)
          {
@@ -778,7 +791,7 @@ DLL_HEADER void ExportNgOCCShapes(py::module &m)
            PropagateProperties(builder, shape, occ2ng(trafo));
            return builder.Shape();
          }, py::arg("axes"),
-         "copy shape, and mirror around axis 'axis'")
+         "copy shape, and rotate by 180 deg around axis 'axis'")
     
     .def("Scale", [](const TopoDS_Shape & shape, const gp_Pnt p, double s)
          {
@@ -815,6 +828,8 @@ DLL_HEADER void ExportNgOCCShapes(py::module &m)
           return nullopt;
       }, [](const TopoDS_Shape & self, optional<string> name) {
         OCCGeometry::GetProperties(self).name = name;
+        for (auto & s : GetHighestDimShapes(self))
+          OCCGeometry::GetProperties(s).name = name;
       }, "'name' of shape")
     
     .def_property("maxh",
@@ -824,12 +839,9 @@ DLL_HEADER void ExportNgOCCShapes(py::module &m)
                   },
                   [](TopoDS_Shape& self, double val)
                   {
-                    for (auto typ : { TopAbs_SOLID, TopAbs_FACE,  TopAbs_EDGE, TopAbs_VERTEX })
-                      for (TopExp_Explorer e(self, typ); e.More(); e.Next())
-                      {
-                        auto & maxh = OCCGeometry::GetProperties(e.Current()).maxh;
-                        maxh = min2(val, maxh);
-                      }
+                    OCCGeometry::GetProperties(self).maxh = val;
+                    for(auto & s : GetHighestDimShapes(self))
+                      OCCGeometry::GetProperties(s).maxh = val;
                   }, "maximal mesh-size for shape")
     
     .def_property("hpref",
@@ -839,8 +851,9 @@ DLL_HEADER void ExportNgOCCShapes(py::module &m)
                   },
                   [](TopoDS_Shape& self, double val)
                   {
-                    auto & hpref = OCCGeometry::GetProperties(self).hpref;
-                    hpref = max2(val, hpref);
+                    OCCGeometry::GetProperties(self).hpref = val;
+                    for(auto & s : GetHighestDimShapes(self))
+                      OCCGeometry::GetProperties(s).hpref = val;
                   }, "number of refinement levels for geometric refinement")
     
     .def_property("col", [](const TopoDS_Shape & self) -> py::object {
@@ -853,6 +866,8 @@ DLL_HEADER void ExportNgOCCShapes(py::module &m)
         if(c.size() == 4)
           col[3] = c[3];
         OCCGeometry::GetProperties(self).col = col;
+        for(auto & s : GetHighestDimShapes(self))
+          OCCGeometry::GetProperties(s).col = col;
       }, "color of shape as RGB - tuple")
     .def_property("layer", [](const TopoDS_Shape& self) {
     if (!OCCGeometry::HaveProperties(self))
@@ -860,6 +875,8 @@ DLL_HEADER void ExportNgOCCShapes(py::module &m)
     return OCCGeometry::GetProperties(self).layer;
     }, [](const TopoDS_Shape& self, int layer) {
     OCCGeometry::GetProperties(self).layer = layer;
+    for(auto & s : GetHighestDimShapes(self))
+      OCCGeometry::GetProperties(s).layer = layer;
     }, "layer of shape")
     .def("UnifySameDomain", [](const TopoDS_Shape& shape,
                                bool edges, bool faces,
@@ -1142,12 +1159,63 @@ DLL_HEADER void ExportNgOCCShapes(py::module &m)
          py::arg("intersection") = false,py::arg("joinType")="arc",
          py::arg("removeIntersectingEdges") = false,
          "makes shell-like solid from faces")
+
+    .def("Offset", [](const TopoDS_Shape & shape, 
+                      double offset, double tol, bool intersection,
+                      string joinT, bool removeIntEdges, optional<string> identification_name) {
+           BRepOffsetAPI_MakeOffsetShape maker;
+           GeomAbs_JoinType joinType;
+           if(joinT == "arc")
+             joinType = GeomAbs_Arc;
+           else if(joinT == "intersection")
+             joinType = GeomAbs_Intersection;
+           else if(joinT == "tangent")
+            joinType = GeomAbs_Tangent;
+           else
+             throw Exception("Only joinTypes 'arc', 'intersection' and 'tangent' exist!");
+           
+           maker.PerformByJoin(shape, offset, tol,
+                               BRepOffset_Skin, intersection,
+                               false, joinType, removeIntEdges);
+
+           // PropagateProperties (maker, shape);
+           for (auto typ : { TopAbs_FACE,  TopAbs_EDGE, TopAbs_VERTEX })
+             for (TopExp_Explorer e(shape, typ); e.More(); e.Next())
+               {
+                 auto s = e.Current();
+                 auto prop = OCCGeometry::GetProperties(s);
+                 for (auto mods : maker.Generated(s))
+                   {
+                     if(OCCGeometry::HaveProperties(s))
+                       {
+                         auto & new_props = OCCGeometry::GetProperties(mods);
+                         new_props.Merge(prop);
+                         if (prop.name) new_props.name = string("offset_")+(*prop.name);
+                       }
+                     if(identification_name)
+                       {
+                         OCCIdentification ident;
+                         ident.from = s;
+                         ident.to = mods;
+                         ident.name = *identification_name;
+                         ident.type = Identifications::CLOSESURFACES;
+                         OCCGeometry::GetIdentifications(s).push_back(ident);
+                       }
+                   }
+               }
+           
+           return maker.Shape();
+       }, py::arg("offset"), py::arg("tol"),
+         py::arg("intersection") = false,py::arg("joinType")="arc",
+         py::arg("removeIntersectingEdges") = false,
+         py::arg("identification_name") = nullopt,
+         "makes shell-like solid from faces")
+
+
     
     .def("MakeTriangulation", [](const TopoDS_Shape & shape)
          {
-           BRepTools::Clean (shape);
-           double deflection = 0.01;
-           BRepMesh_IncrementalMesh (shape, deflection, true);
+           BuildTriangulation(shape);
          })
 
 
@@ -1176,12 +1244,6 @@ DLL_HEADER void ExportNgOCCShapes(py::module &m)
                throw NgException ("Triangulation: shape is not a face");
              }
 
-           /*
-           BRepTools::Clean (shape);
-           double deflection = 0.01;
-           BRepMesh_IncrementalMesh (shape, deflection, true);
-           */
-
            Handle(Geom_Surface) surf = BRep_Tool::Surface (face);
 
            TopLoc_Location loc;
@@ -1189,9 +1251,7 @@ DLL_HEADER void ExportNgOCCShapes(py::module &m)
            
            if (triangulation.IsNull())
              {
-               BRepTools::Clean (shape);
-               double deflection = 0.01;
-               BRepMesh_IncrementalMesh (shape, deflection, true);
+               BuildTriangulation(shape);
                triangulation = BRep_Tool::Triangulation (face, loc);               
              }
            // throw Exception("Don't have a triangulation, call 'MakeTriangulation' first");
@@ -1212,11 +1272,9 @@ DLL_HEADER void ExportNgOCCShapes(py::module &m)
          })
     .def("_webgui_data", [](const TopoDS_Shape & shape)
          {
-           BRepTools::Clean (shape);
-           double deflection = 0.01;
-           BRepMesh_IncrementalMesh (shape, deflection, true);
-           // triangulation = BRep_Tool::Triangulation (face, loc);
-
+           [[maybe_unused]] auto status = BuildTriangulation(shape);
+           // cout << "status = " << aStatus << endl;
+           
            std::vector<double> p[3];
            std::vector<double> n[3];
            py::list names, colors, solid_names;
@@ -1233,20 +1291,14 @@ DLL_HEADER void ExportNgOCCShapes(py::module &m)
                // Handle(TopoDS_Face) face = e.Current();
                fmap.Add(face);
                ExtractFaceData(face, index, p, n, box);
-               auto & props = OCCGeometry::GetProperties(face);
-               if(props.col)
-               {
-                 auto & c = *props.col;
-                 colors.append(py::make_tuple(c[0], c[1], c[2]));
-               }
-               else
-                   colors.append(py::make_tuple(0.0, 1.0, 0.0));
-               if(props.name)
-               {
-                 names.append(*props.name);
-               }
-               else
-                   names.append("");
+
+               ShapeProperties props;
+               if(OCCGeometry::HaveProperties(face))
+                 props = OCCGeometry::GetProperties(face);
+
+               auto c = props.GetColor();
+               colors.append(py::make_tuple(c[0], c[1], c[2], c[3]));
+               names.append(props.GetName());
                index++;
            }
 
@@ -1355,6 +1407,9 @@ DLL_HEADER void ExportNgOCCShapes(py::module &m)
           BRepLib::BuildCurves3d(edge);
           return edge;
         }))
+    .def(py::init([] (const TopoDS_Vertex & v1, const TopoDS_Vertex & v2) {
+      return BRepBuilderAPI_MakeEdge(v1, v2).Edge();
+    }))
     .def("Value", [](const TopoDS_Edge & e, double s) {
         double s0, s1;
         auto curve = BRep_Tool::Curve(e, s0, s1);
@@ -1408,7 +1463,20 @@ DLL_HEADER void ExportNgOCCShapes(py::module &m)
                              return tuple(s0, s1);
                            },
                            "parameter interval of curve")
-    
+    .def_property("partition",
+       [](TopoDS_Shape & self) -> optional<Array<double>>
+       {
+         if (OCCGeometry::HaveProperties(self))
+           return OCCGeometry::GetProperties(self).partition;
+         return nullopt;
+       },
+       [](TopoDS_Shape &self, py::array_t<double> val)
+       {
+         Array<double> partition(val.size());
+         for(auto i : Range(partition))
+           partition[i] = val.at(i);
+         OCCGeometry::GetProperties(self).partition = std::move(partition);
+       })
     
     .def("Split", [](const TopoDS_Edge& self, py::args args)
     {
@@ -1477,6 +1545,24 @@ DLL_HEADER void ExportNgOCCShapes(py::module &m)
               throw NgException("error in wire builder: "+errstr.str());
             }
         }))
+    .def("Offset", [](const TopoDS_Wire & wire, const TopoDS_Face & face, double dist,
+                      string joinT, bool openresult)
+    {
+      GeomAbs_JoinType joinType;
+      if(joinT == "arc")
+        joinType = GeomAbs_Arc;
+      else if(joinT == "intersection")
+        joinType = GeomAbs_Intersection;
+      else if(joinT == "tangent")
+        joinType = GeomAbs_Tangent;
+      else
+        throw Exception("Only joinTypes 'arc', 'tangent', and 'intersection' exist!");
+      BRepOffsetAPI_MakeOffset builder(face, joinType, openresult);
+      builder.AddWire(wire);
+      builder.Perform(dist);
+      auto shape = builder.Shape();    
+      return shape;
+    })
     ;
 
   py::class_<TopoDS_Face, TopoDS_Shape> (m, "Face")
@@ -1589,8 +1675,14 @@ DLL_HEADER void ExportNgOCCShapes(py::module &m)
   
   py::implicitly_convertible<TopoDS_Shape, TopoDS_Face>();
   py::implicitly_convertible<TopoDS_Edge, TopoDS_Wire>();
-  
 
+  m.def("MakePolygon", [](std::vector<TopoDS_Vertex> verts)
+  {
+    BRepBuilderAPI_MakePolygon builder;
+    for(auto& v : verts)
+      builder.Add(v);
+    return builder.Wire();
+  });
 
   class ListOfShapesIterator 
   {
@@ -1740,17 +1832,8 @@ DLL_HEADER void ExportNgOCCShapes(py::module &m)
     },
       [](ListOfShapes& shapes, double maxh)
       {
-        for(auto& shape : shapes)
-          {
-            for(auto& s : GetSolids(shape))
-              OCCGeometry::GetProperties(s).maxh = maxh;
-            for(auto& s : GetFaces(shape))
-              OCCGeometry::GetProperties(s).maxh = maxh;
-            for(auto& s : GetEdges(shape))
-              OCCGeometry::GetProperties(s).maxh = maxh;
-            for(auto& s : GetVertices(shape))
-              OCCGeometry::GetProperties(s).maxh = maxh;
-          }
+        for(auto & s : shapes)
+          OCCGeometry::GetProperties(s).maxh = maxh;
       }, "set maxh for all elements of list")
     .def_property("hpref", [](ListOfShapes& shapes)
     {
@@ -1759,10 +1842,7 @@ DLL_HEADER void ExportNgOCCShapes(py::module &m)
       [](ListOfShapes& shapes, double hpref)
       {
         for(auto& shape : shapes)
-          {
-            auto& val = OCCGeometry::GetProperties(shape).hpref;
-            val = max2(hpref, val);
-          }
+          OCCGeometry::GetProperties(shape).hpref = hpref;
       }, "set hpref for all elements of list")
     .def_property("quad_dominated", [](ListOfShapes& shapes)
                   {
@@ -1935,7 +2015,7 @@ DLL_HEADER void ExportNgOCCShapes(py::module &m)
                     optional<TopoDS_Wire> auxspine) {
           if (twist)
             {
-              auto [pnt, angle] = *twist;
+              // auto [pnt, angle] = *twist;
 
               /*
                 cyl = Cylinder((0,0,0), Z, r=1, h=1).faces[0]
@@ -2230,7 +2310,7 @@ tangents : Dict[int, gp_Vec2d]
       TColgp_Array1OfPnt poles(0, vpoles.size()-1);
       TColStd_Array1OfReal knots(0, vpoles.size()+degree);
       TColStd_Array1OfInteger mult(0, vpoles.size()+degree);
-      int cnt = 0;
+      // int cnt = 0;
 
       for (int i = 0; i < vpoles.size(); i++)
         {
@@ -2555,7 +2635,8 @@ degen_tol : double
     .def("NameVertex", &WorkPlane::NameVertex, py::arg("name"), "name vertex at current position")
     .def("Offset", &WorkPlane::Offset, py::arg("d"), "replace current wire by offset curve of distance 'd'")
     .def("Reverse", &WorkPlane::Reverse, "revert orientation of current wire")
-    .def("Close", &WorkPlane::Close, "draw line to start point of wire, and finish wire")
+    .def("Close", &WorkPlane::Close, py::arg("name")=nullopt,
+         "draw line to start point of wire, and finish wire")
     .def("Finish", &WorkPlane::Finish, "finish current wire without closing")
     .def("Last", &WorkPlane::Last, "(deprecated) returns current wire")
     .def("Wire", &WorkPlane::Last, "returns current wire")
